import hashlib


def passwrod2md5(password: str):
    '''
    Use the MD5 algorithm to encrypt user password

    Args:
        password:

    Returns:

    '''
    if password==None:
        return
    
    m = hashlib.md5(password.encode())
    return m.hexdigest().upper()


if __name__ == '__main__':
    ret = passwrod2md5('root')
    ret = passwrod2md5(ret)
    print(ret.upper())



# from ultipa.types.types import TruncateType
from ultipa.operate.base_extra import BaseExtra
from ultipa.utils import UQLMAKER, CommandList
from ultipa.types import ULTIPA_REQUEST, ULTIPA, ULTIPA_RESPONSE
from ultipa.configuration.RequestConfig import RequestConfig
from ultipa.structs import DBType

JSONSTRING_KEYS = ["graph_privileges", "system_privileges", "policies", "policy", "privilege"]
formatdata = ['graph_privileges']


class TruncateExtra(BaseExtra):
    
	'''
        Processing class that defines settings for advanced operations on graphset.
	'''

	def truncate(self, request: ULTIPA_REQUEST.Truncate,
				 requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.UltipaResponse:
		'''
		Truncate graphset.

		Args:
			request: An object of Truncate class

			requestConfig: An object of RequestConfig class

		Returns:
			Response

		'''
		uqlResponse=ULTIPA_RESPONSE.Response()
		command = CommandList.truncate
		requestConfig.graphName = request.graphSetName

		if request.dbType is not None and request.schema is None:
				uqlResponse.status = ULTIPA.Status(code=ULTIPA.Code.UQL_ERROR, message="schema is missing in the parameters")
				return uqlResponse

		if request.dbType is None and request.schema is not None:
				uqlResponse.status = ULTIPA.Status(code=ULTIPA.Code.UQL_ERROR, message="To truncate schema, DbType is required in the parameters")
				return uqlResponse

		uqlMaker = UQLMAKER(command, commonParams=requestConfig)
		uqlMaker.addParam("graph", request.graphSetName)
		
		if request.dbType is not None:
			if request.dbType == DBType.DBNODE:
					if request.schema:
						if request.schema=="*":
							uqlMaker.addParam("nodes", "*")
						else:	
							uqlMaker.addParam("nodes", "@" + request.schema, notQuotes=True)
			if request.dbType == DBType.DBEDGE:
					if  request.schema:
						if request.schema=="*":
							uqlMaker.addParam("edges", "*")
						else:
							uqlMaker.addParam("edges", "@" + request.schema, notQuotes=True)		

		# if request.all and not request.dbType:
		#     uqlMaker = UQLMAKER(command,commandP=request.dbType,commonParams=requestConfig)

		return self.UqlUpdateSimple(uqlMaker)

	def compact(self, graphName: str,
				requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.Response:
		'''
		Compact graphshet.

		Args:
			graph: The name of graphset

			requestConfig: An object of RequestConfig class

		Returns:
			Response

		'''
		command = CommandList.compact
		uqlMaker = UQLMAKER(command, commonParams=requestConfig)
		uqlMaker.addParam("graph", graphName)
		return self.UqlUpdateSimple(uqlMaker)

	def mountGraph(self, graphName: str,
			  requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.Response:
		'''
		Mount graphshet.

		Args:
			graph: The name of graphset
			
			requestConfig: An object of RequestConfig class

		Returns:
			Response

		'''
		commonP = []
		if graphName:
			commonP = graphName
			requestConfig.graphName = graphName
		uqlMaker = UQLMAKER(command=CommandList.mount, commonParams=requestConfig)
		uqlMaker.setCommandParams(commandP=commonP)
		return self.UqlUpdateSimple(uqlMaker)

	def unmountGraph(self, graphName: str,
				requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.Response:
		'''
		Unmount graphshet.

		Args:
			graph: The name of graphset
			
			requestConfig: An object of RequestConfig class

		Returns:
			Response

		'''
		commonP = []
		if graphName:
			commonP = graphName
			requestConfig.graphName = graphName
		uqlMaker = UQLMAKER(command=CommandList.unmount, commonParams=requestConfig)
		uqlMaker.setCommandParams(commandP=commonP)
		return self.UqlUpdateSimple(uqlMaker)



# -*- coding: utf-8 -*-
# @Time    : 2024/05/14 11:47
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : Privilege.py
from typing import List
from ultipa.structs.BaseModel import BaseModel


class Privilege(BaseModel):
	'''
	    Data class for Privilege.
	'''

	def __init__(self, 	graphPrivileges: List[str],systemPrivileges: List[str]):
		self.graphPrivileges=graphPrivileges
		self.systemPrivileges=systemPrivileges


# -*- coding: utf-8 -*-
# @Time    : 2023/8/1 10:47
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : __init__.py.py
from ultipa.structs.DBType import DBType
from ultipa.structs.BaseModel import BaseModel
from ultipa.structs.Algo import Algo
from ultipa.structs.GraphSet import GraphSet
from ultipa.structs.Edge import Edge
from ultipa.structs.Node import Node
from ultipa.structs.EntityRow import EntityRow
from ultipa.structs.InsertType import InsertType
from ultipa.structs.Property import *
from ultipa.structs.PropertyType import PropertyType
from ultipa.structs.ResultType import ResultType
from ultipa.structs.Schema import Schema
from ultipa.structs.PointType import PointType
from ultipa.structs.Index import Index
from ultipa.structs.User import User
from ultipa.structs.Policy import Policy
from ultipa.structs.Privilege import Privilege
from ultipa.structs.Stats import Stats
from ultipa.structs.Top import Top



# -*- coding: utf-8 -*-
# @Time    : 2023/7/18 11:53
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : uqlHelper.py
from ultipa.utils import CommandList
from ultipa.utils.uql_marker import UQL


class UQLHelper:
	'''
		Class that defines settings for UQL processing.

	'''
	_globalCommand = [
		CommandList.createUser,
		CommandList.showUser,
		CommandList.alterUser,
		CommandList.getUser,
		CommandList.dropUser,
		CommandList.grant,
		CommandList.revoke,
		CommandList.showPolicy,
		CommandList.getPolicy,
		CommandList.createPolicy,
		CommandList.alterPolicy,
		CommandList.dropPolicy,
		CommandList.showPrivilege,
		CommandList.stat,
		CommandList.createGraph,
		CommandList.showGraph,
		CommandList.dropGraph,
		CommandList.alterGraph,
		CommandList.top,
		CommandList.kill,
		CommandList.createBackup,
		CommandList.showBackup,
		CommandList.restoreBackup,
		# CommandList.mount,
		# CommandList.unmount
	]
	_ForceMasterCommands=[
		CommandList.showBackup,
	]
	_write = [
		"alter",
		"create",
		"drop",
		"grant",
		"revoke",
		"LTE",
		"UFE",
		"truncate",
		"compact",
		"insert",
		"upsert",
		"update",
		"delete",
		"clear",
		"stop",
		"pause",
		"resume",
		"top",
		"kill",
		"mount",
		"unmount",
		CommandList.createBackup,
		CommandList.restoreBackup,
	]
	_extra = [
		CommandList.top,
		CommandList.kill,
		CommandList.showTask,
		CommandList.stopTask,
		CommandList.clearTask,
		CommandList.stat,
		CommandList.showGraph,
		CommandList.showAlgo,
		CommandList.createPolicy,
		CommandList.dropPolicy,
		CommandList.showPolicy,
		CommandList.getPolicy,
		CommandList.grant,
		CommandList.revoke,
		CommandList.showPrivilege,
		CommandList.showUser,
		CommandList.getSelfInfo,
		CommandList.createUser,
		CommandList.alterUser,
		CommandList.dropUser,
		CommandList.showIndex,
		# CommandList.clearTask,
	]

	def __init__(self, uql: str):
		self.uql = uql
		self.parseRet = UQL.parse(uql)

	def uqlIsGlobal(self):
		'''
		Judge whether the UQL is Global.

		Args:
			uql: a uql statement

		Returns:

		'''

		if self.parseRet != None:
			c1 = self.parseRet.getFirstCommands()
			c2 = f"{c1}().{self.parseRet.getSecondCommands()}"
			return c1 in UQLHelper._globalCommand or c2 in UQLHelper._globalCommand
		return False


	@staticmethod
	def uqlIsWrite(uql: str):
		'''
		Judge whether the UQL is a write operation.

		Args:
			uql: a uql statement

		Returns:

		'''

		p = UQL.parse(uql)
		if p != None:
			for command in p.commands:
				if list(filter(lambda x: x == command, UQLHelper._write)):
					return True
		return False

	@staticmethod
	def uqlIsForceUseMaster(uql: str):
		p = UQL.parse(uql)
		if p != None:
			for command in p.commands:
				if list(filter(lambda x: x == command, UQLHelper._ForceMasterCommands)):
					return True
		return False

	@staticmethod
	def uqlIsAlgo(uql: str):
		'''
		Judge whether the UQL is an algorithm operation.

		Args:
			uql: a uql statement

		Returns:

		'''
		p = UQL.parse(uql)
		if p != None:
			for command in p.commands:
				if command == 'algo':
					return True
		return False

	@staticmethod
	def uqlIsExecTask(uql: str):
		return "exec task" in uql.lower()

	@staticmethod
	def uqlIsExtra(uql: str):
		p = UQL.parse(uql)
		if p != None:
			c1 = p.getFirstCommands()
			c2 = f"{c1}().{p.getSecondCommands()}"
			return c1 in UQLHelper._extra or c2 in UQLHelper._extra
		return False


from ultipa.configuration.RequestConfig import RequestConfig
from ultipa.operate.base_extra import BaseExtra
from ultipa.structs import DBType
from ultipa.structs.Property import Property
from ultipa.types import ULTIPA, ULTIPA_RESPONSE
from ultipa.utils import UQLMAKER, CommandList
from ultipa.utils.ResposeFormat import ResponseKeyFormat
from ultipa.utils.propertyUtils import getPropertyTypesDesc
from typing import Tuple

BOOL_KEYS = ["index", "lte"]
REPLACE_KEYS = {
	"name": "propertyName",
	"type": "propertyType",
}


class PropertyExtra(BaseExtra):
	'''
	Processing class that defines settings for property related operations.
	'''

	# def listProperty(self, requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.ResponseListProperty:
	# 	'''
	# 	List all properties.

	# 	Args:
	# 		requestConfig: An object of RequestConfig class

	# 	Returns:
	# 		ResponseListProperty

	# 	'''
	# 	return self.showProperty(requestConfig)

	def showProperty(self, dbType: DBType, schemaName:str=None , requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.ResponseListProperty:
		'''
		Show all Node or Edge Schema properties.

		Args:
			dbType: The DBType of data (DBNODE or DBEDGE)

			schema: The name of schema

			requestConfig: An object of RequestConfig class

		Returns:
			ResponseListProperty

		'''
		if dbType != None:
			if dbType == DBType.DBNODE:
				command = CommandList.showNodeProperty
			elif dbType == DBType.DBEDGE:
				command = CommandList.showEdgeProperty
			else:
				command = CommandList.showNodeProperty
			if schemaName:
				commandp = ['@' + f"`{schemaName}`"]
			else:
				commandp = ''
				
		uqlMaker = UQLMAKER(command=command, commonParams=requestConfig)
		uqlMaker.setCommandParams(commandP=commandp)
		# res = self.UqlListSimple(uqlMaker=uqlMaker,
		# 						 responseKeyFormat=ResponseKeyFormat(boolKeys=BOOL_KEYS),
		# 						 isSingleOne=False)
		res=self.uqlSingle(uqlMaker=uqlMaker)
		if dbType==DBType.DBNODE:
			res.items=res.alias("_nodeProperty").asProperties()
		else:
			res.items=res.alias("_edgeProperty").asProperties()	
		return res
	

	def showNodeProperty(self, schemaName:str=None , requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.ResponseListProperty:
		'''
		Show all Node Schema properties.

		Args:
			schemaName: The name of schema

			requestConfig: An object of RequestConfig class

		Returns:
			ResponseListProperty

		'''
		res=self.showProperty(schemaName=schemaName,dbType=DBType.DBNODE,requestConfig=requestConfig)
		return res
	
	def showEdgeProperty(self, schemaName:str=None , requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.ResponseListProperty:
		'''
		Show all Edge Schema properties.

		Args:
			schemaName: The name of schema

			requestConfig: An object of RequestConfig class

		Returns:
			ResponseListProperty

		'''
		
		res=self.showProperty(schemaName=schemaName,dbType=DBType.DBEDGE,requestConfig=requestConfig)
		return res


	def createProperty(self, dbType: DBType, schemaName: str, prop: Property,
					   requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.UltipaResponse:
		'''
		Create a property.

		Args:
			dbType: The DBType of data (DBNODE or DBEDGE)

			schema: The name of schema

			prop:  An object of Property class

			requestConfig: An object of RequestConfig class

		Returns:
			UltipaResponse

		'''
		command = dbType == DBType.DBNODE and CommandList.createNodeProperty or CommandList.createEdgeProperty
		commandP = ["@" + f"`{schemaName}`", f"`{prop.name}`",
					getPropertyTypesDesc(prop.type, prop.subTypes)]

		if prop.description:
			commandP.append(prop.description)
		uqlMaker = UQLMAKER(command=command, commonParams=requestConfig)
		uqlMaker.setCommandParams(commandP=commandP)
		res = self.uqlSingle(uqlMaker)
		return res

	def dropProperty(self, dbType: DBType, schemaName: str, propertyName: str,
					 requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.UltipaResponse:
		'''
		Drop a property.

		Args:
			dbType: The DBType of data (DBNODE or DBEDGE)

			schema: The name of schema

			property: The name of property

			requestConfig: An object of RequestConfig class

		Returns:
			UltipaResponse

		'''
		command = dbType == DBType.DBNODE and CommandList.dropNodeProperty or CommandList.dropEdgeProperty
		commandP = "@`%s`.`%s`" % (schemaName, propertyName)
		uqlMaker = UQLMAKER(command=command, commonParams=requestConfig)
		uqlMaker.setCommandParams(commandP=commandP)
		res = self.uqlSingle(uqlMaker)
		return res

	def alterProperty(self, dbType: DBType, property:Property, newProperty:Property,
					  requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.UltipaResponse:
		'''
		Alter a property.

		Args:
			dbType: The DBType of data (DBNODE or DBEDGE)

			property: The name of property

			newProperty: The new name of property

			requestConfig: An object of RequestConfig class

		Returns:
			UltipaResponse

		'''
		command = dbType == DBType.DBNODE and CommandList.alterNodeProperty or CommandList.alterEdgeProperty
		commandP = "@`%s`.`%s`" % (property.schema, property.name)
		update_dict = {}
		if newProperty.name:
			update_dict.setdefault('name', newProperty.name)
		if newProperty.description:
			update_dict.update({'description': newProperty.description})
		uqlMaker = UQLMAKER(command=command, commonParams=requestConfig)
		uqlMaker.setCommandParams(commandP)
		uqlMaker.addParam("set", update_dict)
		res = self.uqlSingle(uqlMaker)
		return res
	



		



import time

import json
from typing import Tuple

from ultipa.configuration.RequestConfig import RequestConfig
from ultipa.operate.base_extra import BaseExtra
from ultipa.structs import GraphSet
from ultipa.types import ULTIPA, ULTIPA_RESPONSE
from ultipa.utils import UQLMAKER, CommandList
from ultipa.utils.ResposeFormat import ResponseKeyFormat
from ultipa.utils.convert import convertToGraph

REPLACE_KEYS = {
	"graph": "name",
}


class GraphExtra(BaseExtra):

	'''
	Processing class that defines settings for graphset related operations.
	'''

	def uqlCreateSubgraph(self, uql: str, subGraphName: str, requestConfig: RequestConfig = RequestConfig()):
		ret = self.uql(uql, requestConfig)
		graphRet = self.getGraph(subGraphName)
		if graphRet.status.code != 0:
			self.createGraph(GraphSet(name=subGraphName))
			time.sleep(3)

	def showGraph(self, requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.ResponseListGraph:
		'''
		Acquire graphset list.

		Args:
			requestConfig: An object of RequestConfig class

		Returns:
			ResponseListGraph
		'''

		uqlMaker = UQLMAKER(command=CommandList.showGraph, commonParams=requestConfig)
		uqlMaker.setCommandParams("")
		res = self.UqlListSimple(uqlMaker=uqlMaker, responseKeyFormat=ResponseKeyFormat(keyReplace=REPLACE_KEYS))
		res.data=convertToGraph(res)
		return res

	def getGraph(self, graphName: str,
				 requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.ResponseGraph:
		'''
		Acquire a designated graphset.

		Args:
			graphName: The name of graphset

			requestConfig: An object of RequestConfig class

		Returns:
			UltipaResponse
		'''

		uqlMaker = UQLMAKER(command=CommandList.showGraph, commonParams=requestConfig)
		uqlMaker.setCommandParams(graphName)

		res = self.UqlListSimple(uqlMaker=uqlMaker, responseKeyFormat=ResponseKeyFormat(keyReplace=REPLACE_KEYS))
		res.data=convertToGraph(res)
		return res

	def createGraph(self, graph: GraphSet,
					requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.UltipaResponse:
		'''
		Create a graphset.

		Args:
			grpah: An object of Graph class

			requestConfig: An object of RequestConfig class

		Returns:
			UltipaResponse
		'''

		uqlMaker = UQLMAKER(command=CommandList.createGraph, commonParams=requestConfig)
		if graph.description:
			uqlMaker.setCommandParams([graph.name, graph.description])
		else:
			uqlMaker.setCommandParams(graph.name)
		res = self.uqlSingle(uqlMaker)
		return res

	def dropGraph(self, graphName: str,
				  requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.UltipaResponse:
		'''
		Drop a graphset.

		Args:
			graphName: The name of graphset

			requestConfig: An object of RequestConfig class

		Returns:
			UltipaResponse
		'''

		uqlMaker = UQLMAKER(command=CommandList.dropGraph, commonParams=requestConfig)
		uqlMaker.setCommandParams(graphName)
		res = self.uqlSingle(uqlMaker)
		return res

	def alterGraph(self, oldGraph: GraphSet, newGraph: GraphSet,
				   requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.UltipaResponse:
		'''
		Alter a graphset

		Args:
			oldGraphName: The orignal name of graphset

			newGraphName: The new name of graphset

			newDescription: The new description of graphset

			requestConfig: An object of RequestConfig class

		Returns:
			UltipaResponse
		'''

		requestConfig.graphName = oldGraph.name
		uqlMaker = UQLMAKER(command=CommandList.alterGraph, commonParams=requestConfig)
		uqlMaker.setCommandParams(oldGraph.name)
		data = {"name": newGraph.name}
		if newGraph.description is not None:
			data.update({'description': newGraph.description})
		uqlMaker.addParam("set", data)
		res = self.uqlSingle(uqlMaker)
		return res

	def hasGraph(self,graphName:str,requestConfig: RequestConfig = RequestConfig()) -> bool:
		'''
		Check if graph exists or not.

		Args:
			graphName: The name of graphset

			requestConfig: An object of RequestConfig class

		Returns:
			bool
		'''

		graphsdata=self.showGraph(requestConfig).data
		for graphs in graphsdata:
			if(graphs.name==graphName):
				return True
			
		return False	
	
	def createGraphIfNotExist(self, graph: GraphSet,
					requestConfig: RequestConfig = RequestConfig()) -> Tuple[bool,ULTIPA_RESPONSE.UltipaResponse] :
		'''
		Checks if graph exists or not, if graph does not exist then creates new.

		Args:
			graphName: The name of graphset

			requestConfig: An object of RequestConfig class

		Returns:
			bool
			UltipaResponse
		'''

		
		if(self.hasGraph(graph.name,requestConfig))==True:
			return [True,ULTIPA_RESPONSE.UltipaResponse()]
		
		else:
			res=self.createGraph(graph,requestConfig)
			return [False,res]
		






# -*- coding: utf-8 -*-
# @Time    : 2023/7/18 09:20
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : connectionBase.py

import logging
import time
import schedule

from ultipa import ParameterException
from ultipa.configuration.UltipaConfig import UltipaConfig
from ultipa.configuration.RequestConfig import RequestConfig
from ultipa.connection.clientInfo import ClientInfo
from ultipa.connection.clientType import ClientType
from ultipa.connection.hostManagerControl import HostManagerControl, RAFT_GLOBAL
from ultipa.connection.uqlHelper import UQLHelper
from ultipa.proto import ultipa_pb2
from ultipa.utils import CommandList
from ultipa.utils.common import GETLEADER_TIMEOUT
from ultipa.utils.format import FormatType
from ultipa.utils.logger import LoggerConfig
from ultipa.types import ULTIPA, ULTIPA_RESPONSE
from ultipa.utils.ultipaSchedule import run_continuously


class ConnectionBase:
	'''
		A base class that defines settings for an Ultipa connection.

	'''

	def __init__(self, host: str, defaultConfig: UltipaConfig, crtFilePath: str = None):
		self.host = host
		self.username = defaultConfig.username
		self.password = defaultConfig.password
		self.crtPath = crtFilePath
		self.defaultConfig = defaultConfig
		self.runSchedule: object = None
		self.crt = None
		if crtFilePath:
			try:
				with open(f'{crtFilePath}', 'rb') as f:
					self.crt = f.read()
			except Exception as e:
				raise ParameterException(err=e)
		self.hostManagerControl = HostManagerControl(self.host, self.username, self.password,
													 self.defaultConfig.maxRecvSize, self.crt,
													 consistency=defaultConfig.consistency)

		self.defaultConfig.defaultGraph = defaultConfig.defaultGraph or "default"
		self.defaultConfig.timeoutWithSeconds = defaultConfig.timeoutWithSeconds or 15
		self.defaultConfig.responseWithRequestInfo = defaultConfig.responseWithRequestInfo or False
		self.defaultConfig.consistency = defaultConfig.consistency
		self.graphSetName = self.defaultConfig.defaultGraph
		self.count = 0

		if not self.defaultConfig.uqlLoggerConfig and self.defaultConfig.Debug:
			self.defaultConfig.uqlLoggerConfig = LoggerConfig(name="ultipa", fileName="",
															  isStream=self.defaultConfig.Debug, isWriteToFile=False,
															  level=logging.INFO)

	def getGraphSetName(self, currentGraphName: str, uql: str = "", isGlobal: bool = False):
		if isGlobal:
			return RAFT_GLOBAL
		if uql:
			parse = UQLHelper(uql)
			if parse.uqlIsGlobal():
				return RAFT_GLOBAL
			c1 = parse.parseRet.getFirstCommands()
			c2 = f"{c1}().{parse.parseRet.getSecondCommands()}"
			if c2 in [CommandList.mount, CommandList.unmount, CommandList.truncate]:
				graphName = parse.parseRet.getCommandsParam(1)
				if graphName:
					return graphName
		return currentGraphName or self.defaultConfig.defaultGraph

	def getTimeout(self, timeout: int):
		return timeout or self.defaultConfig.timeoutWithSeconds

	def getClientInfo(self, clientType: int = ClientType.Default, graphSetName: str = '', uql: str = '',
					  isGlobal: bool = False, ignoreRaft: bool = False, useHost: str = None, useMaster: bool = False,
					  timezone=None, timeZoneOffset=None):
		goGraphName = self.getGraphSetName(currentGraphName=graphSetName, uql=uql, isGlobal=isGlobal)
		if not ignoreRaft and not self.hostManagerControl.getHostManger(goGraphName).raftReady:
			refreshRet = self.refreshRaftLeader(self.hostManagerControl.initHost,
												RequestConfig(graphName=goGraphName))
			self.hostManagerControl.getHostManger(goGraphName).raftReady = refreshRet

		clientInfo = self.hostManagerControl.chooseClientInfo(type=clientType, uql=uql, graphSetName=goGraphName,
															  useHost=useHost, useMaster=useMaster)
		metadata = clientInfo.getMetadata(goGraphName, timezone, timeZoneOffset)
		return ClientInfo(Rpcsclient=clientInfo.Rpcsclient, Controlsclient=clientInfo.Controlsclient, metadata=metadata,
						  graphSetName=goGraphName, host=clientInfo.host)

	def getRaftLeader(self, requestConfig: RequestConfig = RequestConfig()):
		resRaftLeader = self.__autoGetRaftLeader(host=self.host, requestConfig=requestConfig)
		RaftStatus = FormatType.getRaftStatus(resRaftLeader)
		return ULTIPA_RESPONSE.Response(RaftStatus)

	def __getRaftLeader(self, requestConfig: RequestConfig = RequestConfig()):
		if requestConfig == None:
			graphSetName = None
		else:
			if not requestConfig.graphName:
				graphSetName = 'default'
			else:
				graphSetName = requestConfig.graphName

		clientInfo = self.getClientInfo(clientType=ClientType.Leader, graphSetName=graphSetName, ignoreRaft=True,
										useMaster=requestConfig.useMaster)
		res = clientInfo.Controlsclient.GetLeader(ultipa_pb2.GetLeaderRequest(), metadata=clientInfo.metadata,
												  timeout=GETLEADER_TIMEOUT)

		return FormatType.Response(_res=res, host=self.host)

	def __autoGetRaftLeader(self, host: str, requestConfig: RequestConfig, retry=0):
		'''For internal use, return customized value'''
		conn = ConnectionBase(host=host, crtFilePath=self.crtPath, defaultConfig=self.defaultConfig)
		try:
			res = conn.__getRaftLeader(requestConfig)
		except Exception as e:
			self.hostManagerControl.initHost = conn.host
			if host in self.defaultConfig.hosts:
				self.defaultConfig.hosts.remove(host)
			return {
				"code": ULTIPA.Code.FAILED,
				"message": str(e._state.code) + ' : ' + str(e._state.details)
			}
		status = res.status
		if status.code == ULTIPA.Code.SUCCESS:
			# status.clusterInfo.raftPeers.remove(host)  # remove leader
			self.hostManagerControl.initHost = host
			for i in status.clusterInfo.raftPeers:
				if i.host == host:
					status.clusterInfo.raftPeers.remove(i)
			return {
				"code": status.code,
				"message": status.message,
				'leaderHost': host,
				"followersPeerInfos": list(filter(lambda x: x != host, status.clusterInfo.raftPeers)),
				"leaderInfos": status.clusterInfo.leader,
			}
		elif status.code == ULTIPA.Code.NOT_RAFT_MODE:
			return {
				"code": status.code,
				"message": status.message,
				"leaderHost": host,
				"followersPeerInfos": [],
				"leaderInfos": status.clusterInfo.leader
			}
		elif status.code in [ULTIPA.Code.RAFT_REDIRECT, ULTIPA.Code.RAFT_LEADER_NOT_YET_ELECTED,
							 ULTIPA.Code.RAFT_NO_AVAILABLE_FOLLOWERS, ULTIPA.Code.RAFT_NO_AVAILABLE_ALGO_SERVERS]:
			if retry > 2:
				return {
					"code": status.code,
					"message": status.message,
					"redirectHost": res.status.clusterInfo.redirect
				}
			if status.code != ULTIPA.Code.RAFT_REDIRECT:
				time.sleep(0.3)
			if status.code == ULTIPA.Code.RAFT_REDIRECT:
				host = res.status.clusterInfo.redirect
			return self.__autoGetRaftLeader(host=host, requestConfig=requestConfig, retry=retry + 1)

		return {
			"code": status.code,
			"message": status.message
		}

	def refreshRaftLeader(self, redirectHost: str, requestConfig: RequestConfig):
		# hosts = [redirectHost] if redirectHost else self.hostManagerControl.getAllHosts()
		hosts = [redirectHost] if redirectHost else []
		goGraphName = self.getGraphSetName(requestConfig.graphName)

		for h in self.defaultConfig.hosts:
			if h not in hosts:
				hosts.append(h)

		# print("host:", hosts)
		for host in hosts:
			resRaftLeader = self.__autoGetRaftLeader(host=host, requestConfig=requestConfig)
			code = resRaftLeader["code"]
			if code == ULTIPA.Code.SUCCESS:
				leaderHost = resRaftLeader["leaderHost"]
				followersPeerInfos = resRaftLeader["followersPeerInfos"]
				leaderInfos = resRaftLeader["leaderInfos"]
				hostManager = self.hostManagerControl.upsetHostManger(goGraphName, leaderHost)
				hostManager.setClients(leaderHost=leaderHost, followersPeerInfos=followersPeerInfos,
									   leaderInfos=leaderInfos)
				return True
		# elif code == ULTIPA.Code.RAFT_REDIRECT:
		# 	return False
		return False

	def stopConnectionAlive(self):
		if self.runSchedule != None:
			self.runSchedule.set()

	def keepConnectionAlive(self, timeIntervalSeconds: int = None):
		timeIntervalSeconds = self.defaultConfig.heartBeat if timeIntervalSeconds == None else timeIntervalSeconds

		def test_allconn():
			goGraphName = self.defaultConfig.defaultGraph
			for host in self.hostManagerControl.getAllClientInfos(goGraphName):
				res = host.Controlsclient.SayHello(ultipa_pb2.HelloUltipaRequest(name="test"),
												   metadata=host.getMetadata(goGraphName, None, None))
				# print(host.host,res.message)
				if self.defaultConfig.uqlLoggerConfig is None:
					self.defaultConfig.uqlLoggerConfig = LoggerConfig(name="HeartBeat", fileName=None,
																	  isWriteToFile=False, level=logging.WARN,
																	  isStream=True)
				self.defaultConfig.uqlLoggerConfig.getlogger().info(f"HeartBeat:{host.host}--{res.message}")

		schedule.every().second.do(test_allconn)
		self.runSchedule = run_continuously(timeIntervalSeconds)



# -*- coding: utf-8 -*-
# @Time    : 2023/8/10 18:25
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : UltipaConfig.py
from typing import List
from ultipa.utils.logger import LoggerConfig
from enum import Enum
from ultipa.utils.password2md5 import passwrod2md5


class PasswordEncrypt(Enum):
    MD5 = 1
    LDAP = 2
    NOTHING = 3

class UltipaConfig:
	'''
    Configuration class used to instantiate Ultipa connection.

    This class stores settings for establishing a connection to an Ultipa server. The change of an UltipaConfig object will update the connection it established.
    
    Args:
        - hosts (str): The list of Ultipa server hosts.
        - username (list[str]): The username of server.
        - password (str): The password of server.
		- passwordEncrypt(PasswordEncrypt): Encryption type for password(MD5/LDAP/NOTHING)
        - crtFilePath (str): The file path of SSL certificate when both Ultipa server and client-end are in SSL mode.
        - defaultGraph (str): The name of graphset to use.
        - timeout (int): The timeout for the request in seconds.
        - responseWithRequestInfo (bool): Whether to return request.
        - consistency (bool): Whether to use leader host to guarantee Consistency Read.
        - heartBeat (int): The heartbeat in seconds for all instances, set 0 to turn off heartbeat.
        - maxRecvSize (int): The maximum number of bytes when receiving data.
        - uqlLoggerConfig (LoggerConfig): The object of LoggerConfig class.
        - debug (bool): Whether to use debug mode.
        - timeZone (str): The string of timezone in standard format.
        - timeZoneOffset (any): 1, the number of seconds; 2, a 5-character string such as +0700, -0430.
        
    To enable debug mode, a logger configuration is necessary. If 'debug' is set to True but 'uqlLoggerConfig' is not defined, a LoggerConfig object will be generated for debugging purposes; however, the log will not be written to a file.
	'''

	def __init__(self, hosts:List[str]=None, username:str=None, password:str=None, passwordEncrypt:PasswordEncrypt=PasswordEncrypt.MD5, crtFilePath:str=None, defaultGraph: str = "default",
				 timeout: int = 3600, responseWithRequestInfo: bool = False,
				 consistency: bool = False, heartBeat: int = 10, maxRecvSize: int = -1,
				 uqlLoggerConfig: LoggerConfig = None, debug: bool = False, timeZone=None,
				 timeZoneOffset=None, **kwargs):
		if hosts is None:
			hosts = []
		self.hosts = hosts
		self.username = username
		# self.password = password
		self._password = None
		self.passwordEncrypt=passwordEncrypt
		self.crtFilePath = crtFilePath
		self.defaultGraph = defaultGraph
		self.timeoutWithSeconds = timeout
		self.responseWithRequestInfo = responseWithRequestInfo
		# Read consistency, when set to False, distributes the workload across nodes.
		self.consistency = consistency
		self.uqlLoggerConfig = uqlLoggerConfig
		self.heartBeat = heartBeat
		self.maxRecvSize = maxRecvSize
		self.Debug = debug
		self.timeZone = timeZone
		self.timeZoneOffset = timeZoneOffset
		if kwargs.get("timeoutWithSeconds") is not None:
			self.timeoutWithSeconds = kwargs.get("timeoutWithSeconds")
		
		self.password = password 
	@property
	def password(self):
			return self._password

	@password.setter
	def password(self, value):
			if self.passwordEncrypt == PasswordEncrypt.MD5 and value is not None:
				self._password = passwrod2md5(value)
			else:
				self._password = value			

		# if self.passwordEncrypt==PasswordEncrypt.MD5 and password is not None:
		# 	passworde=passwrod2md5(self.password)
		# 	self.password=passworde
			
	def setDefaultGraphName(self, graph: str):
		self.defaultGraph = graph


# -*- coding: utf-8 -*-
# @Time    : 2023/8/1 10:47
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : Graph.py
from ultipa.structs.BaseModel import BaseModel

class GraphSet(BaseModel):
    id: str
    name: str
    totalNodes: str
    totalEdges: str
    description: str
    status: str

    def __init__(self, name: str, id: str = None, totalNodes: str = None, 
                 totalEdges: str = None, description: str = None, status: str = None):
        self.id = id
        self.name = name
        self.totalNodes = totalNodes
        self.totalEdges = totalEdges
        self.description = description
        self.status = status






# -*- coding: utf-8 -*-
# @Time    : 2024/05/14 11:47
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : User.py
from typing import List
from ultipa.structs.BaseModel import BaseModel


class User(BaseModel):
	'''
	    Data class for User.
	'''

	def __init__(self, username: str, create:str=None,graphPrivileges: dict=None,	systemPrivileges: List[str]=None,policies: List[str]=None,propertyPrivileges:dict=None):
		self.username = username
		self.create = create
		self.graphPrivileges = graphPrivileges
		self.systemPrivileges = systemPrivileges
		self.policies = policies
		self.propertyPrivileges=propertyPrivileges


from ultipa.operate.base_extra import BaseExtra
from ultipa.structs import DBType
from ultipa.types import ULTIPA_REQUEST, ULTIPA, ULTIPA_RESPONSE
from ultipa.utils import UQLMAKER, CommandList
from ultipa.structs.Property import Property
from ultipa.configuration.RequestConfig import RequestConfig

class LteUfeExtra(BaseExtra):

	'''
	Processsing class that defines settings for LTE and UFE related operations.
	'''

	def lte(self, dbType:DBType, propertyName:str,schemaName:str=None, 
			requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.UltipaResponse:
		'''
		Load properties to memory (LTE).

		Args:
			dbType: The DBType of data (DBNODE or DBEDGE)

			schemaName: The name of schema

			property: An object of Property class

			requestConfig: An object of RequestConfig class

		Returns:
			UltipaResponse
		'''

		command = dbType == DBType.DBNODE and CommandList.lteNode or CommandList.lteEdge
		uqlMaker = UQLMAKER(command=command, commonParams=requestConfig)
		if schemaName:
			commandP = "@`%s`.`%s`" % (schemaName, propertyName)
		elif schemaName==None:
			commandP="@`*`.`%s`" %(propertyName)	
		uqlMaker.setCommandParams(commandP=commandP)
		res = self.uqlSingle(uqlMaker=uqlMaker)
		return res

	def ufe(self, dbType:DBType, propertyName:str, schemaName:str=None,
			requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.UltipaResponse:
		'''
		Unload properties from memory (UFE).

		Args:
			dbType: The DBType of data (DBNODE or DBEDGE)

			schemaName: The name of schema

			property: An object of Property class

			requestConfig: An object of RequestConfig class

		Returns:
			UltipaResponse
		'''

		command = dbType == DBType.DBNODE and CommandList.ufeNode or CommandList.ufeEdge
		uqlMaker = UQLMAKER(command=command, commonParams=requestConfig)
		if schemaName:
			commandP = "@`%s`.`%s`" % (schemaName, propertyName)
		elif schemaName==None:
			commandP="@`*`.`%s`" %(propertyName)	
		uqlMaker.setCommandParams(commandP=commandP)
		res = self.uqlSingle(uqlMaker=uqlMaker)
		return res



from ultipa.configuration.RequestConfig import RequestConfig
from ultipa.operate.base_extra import BaseExtra
from ultipa.types import ULTIPA_REQUEST, ULTIPA_RESPONSE, ULTIPA
from ultipa.utils import UQLMAKER, CommandList
from ultipa.utils.errors import ParameterException
from ultipa.utils.ResposeFormat import ResponseKeyFormat
from typing import List
from ultipa.structs.Policy import Policy
from ultipa.utils.convert import convertToPolicy
from ultipa.utils.convert import convertToPrivilege


JSONSTRING_KEYS = ["graphPrivileges", "systemPrivileges", "policies", "policy", "privilege"]
formatdata = ['graph_privileges']


class PolicyExtra(BaseExtra):

	'''
		Processing class that defines settings for policy related operations.
	'''

	def showPolicy(self,
				   requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.ResponseListPolicy:
		'''
		Show policy list.

		Args:
			requestConfig: An object of RequestConfig class

		Returns:
			ResponsePolicy

		'''
		uqlMaker = UQLMAKER(command=CommandList.showPolicy, commonParams=requestConfig)
		res = self.UqlListSimple(uqlMaker=uqlMaker,
								 responseKeyFormat=ResponseKeyFormat(jsonKeys=JSONSTRING_KEYS, dataFormat=formatdata))
		if len(res.data)>0:
			res.data=convertToPolicy(res)
		return res

	def showPrivilege(self,
					  requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.ResponsePrivilege:
		'''
		Show privilege list.

		Args:
			requestConfig: An object of RequestConfig class

		Returns:
			ResponsePrivilege

		'''

		uqlMaker = UQLMAKER(command=CommandList.showPrivilege, commonParams=requestConfig)
		res = self.UqlListSimple(uqlMaker=uqlMaker,
								 responseKeyFormat=ResponseKeyFormat(jsonKeys=JSONSTRING_KEYS))
		if len(res.data)>0:
			res.data=convertToPrivilege(res)
		return res

	def getPolicy(self, policyName:str,
				  requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.ResponsePolicy:
		'''
		Get a policy.

		Args:
			name: The name of policy

			requestConfig: An object of RequestConfig class

		Returns:
			ResponsePolicy

		'''

		uqlMaker = UQLMAKER(command=CommandList.getPolicy, commonParams=requestConfig)
		uqlMaker.setCommandParams(policyName)
		res = self.UqlListSimple(uqlMaker=uqlMaker,
								 responseKeyFormat=ResponseKeyFormat(jsonKeys=JSONSTRING_KEYS, dataFormat=formatdata))
		# if isinstance(res.data, list) and len(res.data) > 0:
		# 	res.data = res.data[0]
		if res.status.code==ULTIPA.Code.SUCCESS:
			res.data=convertToPolicy(res)
		return res

	def createPolicy(self, policy: Policy,
					 requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.UltipaResponse:
		'''
		Create a policy.

		Args:
			request:  An object of Policy class

			requestConfig: An object of RequestConfig class

		Returns:
			UltipaResponse

		'''

		uqlMaker = UQLMAKER(command=CommandList.createPolicy, commonParams=requestConfig)
		paramsP = [policy.name]
		if policy.graphPrivileges:
			paramsP.append(policy.graphPrivileges)
		else:
			paramsP.append({})

		if policy.systemPrivileges:
			paramsP.append(policy.systemPrivileges)
		else:
			paramsP.append([])

		if policy.policies:
			paramsP.append(policy.policies)	
		else:
			paramsP.append([])

		if policy.propertyPrivileges:
			paramsP.append(policy.propertyPrivileges)
		else:
			paramsP.append({})	

		uqlMaker.setCommandParams(paramsP)
		res = self.uqlSingle(uqlMaker=uqlMaker)
		return res

	def alterPolicy(self, policy: Policy,
					requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.UltipaResponse:
		'''
		Alter a policy.

		Args:
			request:  An object of Policy class

			requestConfig: An object of RequestConfig class

		Returns:
			UltipaResponse

		'''

		uqlMaker = UQLMAKER(command=CommandList.alterPolicy, commonParams=requestConfig)
		uqlMaker.setCommandParams(policy.name)
		params = {}
		if policy.systemPrivileges is not None:
			params.update({"system_privileges": policy.systemPrivileges})
		if policy.graphPrivileges is not None:
			params.update({"graph_privileges": policy.graphPrivileges})

		if policy.policies is not None:
			params.update({"policies": policy.policies})	

		if policy.propertyPrivileges is not None:
			params.update({"property_privileges": policy.propertyPrivileges})

			
		uqlMaker.addParam('set', params, notQuotes=True)
		res = self.uqlSingle(uqlMaker=uqlMaker)
		return res

	def dropPolicy(self, policyName:str,
				   requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.UltipaResponse:
		'''
		Drop a policy.

		Args:
			policyName:  Name of policy to be dropped
			requestConfig: An object of RequestConfig class

		Returns:
			UltipaResponse

		'''

		uqlMaker = UQLMAKER(command=CommandList.dropPolicy, commonParams=requestConfig)
		uqlMaker.setCommandParams(policyName)
		res = self.uqlSingle(uqlMaker=uqlMaker)
		return res

	def grantPolicy(self, username: str = '', graphPrivileges: dict = None,
				 systemPrivileges: List[str] = None, policies: List[str] = None, propertyPrivileges:dict=None,
					requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.UltipaResponse:
		'''
		Grant privileges and policies to a user.

		Args:
			username (str): The username to grant privileges and policies.

        	graphPrivileges (dict): Dictionary containing graph privileges.

        	systemPrivileges (List[str]): List of system privileges.

        	policies (List[str]): List of policies to grant.
			
			propertyPrivileges(dict): Dictionary containing property privileges

        	requestConfig (RequestConfig): An object of RequestConfig class

		Returns:
			UltipaResponse

		'''

		uqlMaker = UQLMAKER(command=CommandList.grantUser, commonParams=requestConfig)
		if username:
			uqlMaker.setCommandParams(username)
		else:
			raise ParameterException(err='username is a required parameter')

		paramsDict = {}
		if graphPrivileges:
			paramsDict.setdefault('graph_privileges', graphPrivileges)

		if systemPrivileges:
			paramsDict.setdefault('system_privileges',systemPrivileges)

		if policies:
			paramsDict.setdefault('policies',policies)	
		
		if propertyPrivileges:
			paramsDict.setdefault('property_privileges',propertyPrivileges)	


		uqlMaker.addParam('params', paramsDict, notQuotes=True)
		res = self.uqlSingle(uqlMaker=uqlMaker)
		return res

	def revokePolicy(self, username: str = '', graphPrivileges: dict = None,
				 systemPrivileges: List[str] = None, policies: List[str] = None, propertyPrivileges:dict=None,
					 requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.UltipaResponse:
		'''
		Revoke privileges and policies from a user.

		Args:
			username (str): The username to revoke privileges and policies.

        	graphPrivileges (dict): Dictionary containing graph privileges.

        	systemPrivileges (List[str]): List of system privileges.

			propertyPrivileges(dict): Dictionary containing property privileges

        	policies (List[str]): List of policies to revoke.

        	requestConfig (RequestConfig): An object of RequestConfig class

		Returns:
			UltipaResponse

		'''

		uqlMaker = UQLMAKER(command=CommandList.revoke, commonParams=requestConfig)
		if username:
			uqlMaker.setCommandParams(username)
		else:
			raise ParameterException(err='username is a required parameter')

		paramsDict = {}
		if graphPrivileges:
			paramsDict.setdefault('graph_privileges', graphPrivileges)

		if systemPrivileges:
			paramsDict.setdefault('system_privileges', systemPrivileges)

		if propertyPrivileges:
			paramsDict.setdefault('property_privileges',propertyPrivileges)	

		if policies:
			paramsDict.setdefault('policies', policies)	
		uqlMaker.addParam('params', paramsDict, notQuotes=True)
		res = self.uqlSingle(uqlMaker=uqlMaker)
		return res



'''
Settings of some constant values.
'''

LIMIT = 1

DEPTH = 1

GETLEADER_TIMEOUT = 5
# Whether to convert 'dateTime' type to string with format  1970-01-01 08:00:00
DateTimeIs2Str = True
# Whether to convert 'timestamp' type to string with format  1970-01-01 08:00:00  (in timezone of the client end)
TimestampIs2Str = False
# Whether to use 7-digit for the accuracy of 'float' type
Float_Accuracy = True


from typing import List,Callable
from ultipa.operate.base_extra import BaseExtra
from ultipa.proto import ultipa_pb2
from ultipa.types import ULTIPA_REQUEST, ULTIPA_RESPONSE, ULTIPA
from ultipa.utils.format import FormatType
from ultipa.configuration.RequestConfig import RequestConfig
from ultipa.connection.clientType import ClientType
from ultipa.operate.task_extra import TaskExtra

class DownloadExtra(BaseExtra):
	'''
	Processing class that defines settings for file downloading operation.

	'''

	def _downloadAlgoResultFile(self,fileName: str, taskId: str,cb: Callable[[bytes], None],
				 requestConfig: RequestConfig = RequestConfig()):
		'''
		Download file.

		Args:
			filename: Name of the file

			taskID: id of the task

			cb: Callback function that accepts bytes

			requestConfig: An object of RequestConfig class

		Returns:
			stream
		'''
		downResponse = ULTIPA_RESPONSE.Response()
		try:
			
			clientInfo = self.getClientInfo(graphSetName=requestConfig.graphName, useMaster=requestConfig.useMaster,clientType=ClientType.Leader)
			res = clientInfo.Controlsclient.DownloadFileV2(
				ultipa_pb2.DownloadFileRequestV2(file_name=fileName, task_id=taskId),metadata=clientInfo.metadata)			
			
			for data_flow in res:
				ultipa_response = ULTIPA_RESPONSE.Response()
				status = FormatType.status(data_flow.status)
				ultipa_response.status = status
				if status.code != ULTIPA.Code.SUCCESS:						
					cb(ultipa_response)
					break
				ultipa_response.data = data_flow.chunk
				cb(ultipa_response.data)
		except Exception as e:
			downResponse.status = ULTIPA.Status(code=ULTIPA.Code.UNKNOW_ERROR, message=str(e))
			print(downResponse.status.message)
		

	def downloadAlgoResultFile(self,fileName: str, taskId: str,cb:Callable[[bytes], None],
				 requestConfig: RequestConfig = RequestConfig()):
		'''
		Download file.
		
		Args:
			filename: Name of the file

			taskID: id of the task

			cb:Callback function that accepts bytes

			requestConfig: An object of RequestConfig class

		Returns:
			stream
		'''
		return self._downloadAlgoResultFile(fileName=fileName,taskId=taskId,cb=cb,requestConfig=requestConfig)
	

	def _downloadAllAlgoResultFile(self,fileName: str, taskId: str, cb: Callable[[bytes,str], None],
				 requestConfig: RequestConfig = RequestConfig()):
		'''
		Download all files.

		Args:
			filename: Name of the file

			taskID: id of the task

			cb: Callback function that accepts bytes and string inputs

			requestConfig: An object of RequestConfig class

		Returns:
			stream
		'''
		downResponse = ULTIPA_RESPONSE.Response()
		try:
			
			clientInfo = self.getClientInfo(graphSetName=requestConfig.graphName, useMaster=requestConfig.useMaster,clientType=ClientType.Leader)
			res = clientInfo.Controlsclient.DownloadFileV2(
				ultipa_pb2.DownloadFileRequestV2(file_name=fileName, task_id=str(taskId)),metadata=clientInfo.metadata)			
			
			for data_flow in res:
				ultipa_response = ULTIPA_RESPONSE.Response()
				status = FormatType.status(data_flow.status)
				ultipa_response.status = status
				if status.code != ULTIPA.Code.SUCCESS:						
					cb(ultipa_response)
					break
				ultipa_response.data = data_flow.chunk
				cb(ultipa_response.data,fileName)
		except Exception as e:
			downResponse.status = ULTIPA.Status(code=ULTIPA.Code.UNKNOW_ERROR, message=str(e))
			print(downResponse.status.message)
		
	def downloadAllAlgoResultFile(self,taskId: str,cb: Callable[[bytes,str], None],
				 requestConfig: RequestConfig = RequestConfig()):
		'''
		Download all files.

		Args:

			taskID: id of the task

			cb: callback function for receiving data

			requestConfig: An object of RequestConfig class

		Returns:
			stream
		'''
		res=TaskExtra.showTask(self,algoNameOrId=int(taskId))
		file_data=res.data[0].result['result_files']
		names= file_data.split(',')
		for name in names:
			self._downloadAllAlgoResultFile(fileName=name,taskId=taskId,cb=cb,requestConfig=requestConfig)



import logging


class LoggerConfig:
	'''
		A class that defines how log messages related to Ultipa connection are handled. 
		
		It specifies settings such as log name, log file paths, log level, etc.
	'''

	def __init__(self, name: str, fileName: str = None, isWriteToFile: bool = False,
				 level: logging = logging.INFO, isStream: bool = False):
		self.name = name
		self.filename = fileName
		self.isWriteToFile = isWriteToFile
		self.level = level
		self.isStream = isStream

	def getlogger(self):
		fmt = '%(asctime)s - %(levelname)s: %(message)s'
		format_str = logging.Formatter(fmt)
		logger = logging.getLogger(self.name)
		logger.name = self.name
		logger.setLevel(level=self.level)
		if logger.handlers:
			return logger
		if self.isStream:
			sh = logging.StreamHandler()
			sh.setFormatter(format_str)
			logger.addHandler(sh)
		if self.isWriteToFile and self.filename:
			th = logging.FileHandler(filename=self.filename, encoding='utf-8', delay=True)
			th.setFormatter(format_str)
			logger.addHandler(th)
		return logger



# -*- coding: utf-8 -*-
# @Time    : 2023/7/18 17:17
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : base_extra.py
import copy
import csv
import json
import types
from typing import List,Callable

from ultipa import ParameterException
from ultipa.configuration.RequestConfig import RequestConfig
from ultipa.configuration.InsertConfig import InsertConfig
from ultipa.connection.clientType import ClientType
from ultipa.connection.commonUql import GetPropertyBySchema
from ultipa.connection.connectionBase import ConnectionBase
from ultipa.connection.uqlHelper import UQLHelper
from ultipa.proto import ultipa_pb2
from ultipa.structs import DBType
from ultipa.structs.Schema import Schema
from ultipa.types import ULTIPA, ULTIPA_REQUEST, ULTIPA_RESPONSE
from ultipa.types.types_response import PropertyTable
from ultipa.utils import UQLMAKER, CommandList
from ultipa.utils.ResposeFormat import ResponseKeyFormat
from ultipa.utils.convert import convertTableToDict, convertToListAnyObject
from ultipa.utils.format import FormatType
from ultipa.utils.raftRetry import RetryHelp
from ultipa.utils.ultipa_datetime import getTimeZoneOffset, getTimeOffsetSeconds
from ultipa.types.types_response import UQLResponseStream
from ultipa.structs.Node import Node
from ultipa.structs.Edge import Edge
from ultipa.utils.convert import convertToStats


class BaseExtra(ConnectionBase):
	'''
		Processing class that defines settings for basic operations.

	'''

	def test(self,
			 requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.Response:
		'''
		Test connection.

		Args:
			requestConfig: An object of RequestConfig class

		Returns:
			Response
		'''
		testResponse = ULTIPA_RESPONSE.Response()
		returnReq = ULTIPA.ReturnReq(requestConfig.graphName, "test", None, None, False)
		try:
			clientInfo = self.getClientInfo(useHost=requestConfig.useHost, useMaster=requestConfig.useMaster)
			name = 'Test'
			res = clientInfo.Controlsclient.SayHello(ultipa_pb2.HelloUltipaRequest(name=name),
													 metadata=clientInfo.metadata)
			returnReq.host = clientInfo.host
			if (res.message == name + " Welcome To Ultipa!"):
				if self.defaultConfig.uqlLoggerConfig:
					self.defaultConfig.uqlLoggerConfig.getlogger().info(res.message)

				testResponse.status = ULTIPA.Status(code=res.status.error_code, message=res.status.msg)
			else:
				testResponse.status = ULTIPA.Status(code=res.status.error_code, message=res.status.msg)
		except Exception as e:
			testResponse = ULTIPA_RESPONSE.Response()
			try:
				message = str(e._state.details)
			except:
				message = str(e)
			testResponse.status = ULTIPA.Status(code=ULTIPA.Code.UNKNOW_ERROR, message=message)
		if self.defaultConfig.responseWithRequestInfo:
			testResponse.req = returnReq
		return testResponse

	def stats(self,
			  requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.ResponseStat:
		'''
		Query for the server statistics.

		Args:
			requestConfig: An object of RequestConfig class

		Returns:
			ResponseStat
		'''
		uqlMaker = UQLMAKER(command=CommandList.stat, commonParams=requestConfig)
		ret = self.UqlListSimple(uqlMaker=uqlMaker, isSingleOne=True)
		if ret.status.code == ULTIPA.Code.SUCCESS:
			ret.data =convertToStats(ret)
		return ret

	def exportData(self, request: ULTIPA_REQUEST.Export, cb: Callable[[List[Node], List[Edge]],None],
				   requestConfig: RequestConfig = RequestConfig()):
		try:
			req = ultipa_pb2.ExportRequest(db_type=request.dbType, limit=request.limit,
										   select_properties=request.selectPropertiesName, schema=request.schemaName)

			clientInfo = self.getClientInfo(graphSetName=requestConfig.graphName, useMaster=requestConfig.useMaster)
			res = clientInfo.Controlsclient.Export(req, metadata=clientInfo.metadata)
			# res = FormatType.exportResponse(_res=res, timeZone=requestConfig.timeZone,
			# 								timeZoneOffset=requestConfig.timeZoneOffset)				
			nodedata = []
			edgedata = []
			response = ULTIPA_RESPONSE.Response()
			for exportReply in res:
				response.status = FormatType.status(exportReply.status)
				if exportReply.node_table:
					nodedata = FormatType.export_nodes(exportReply, requestConfig.timeZone, requestConfig.timeZoneOffset)
				if exportReply.edge_table:
					edgedata = FormatType.export_edges(exportReply, requestConfig.timeZone, requestConfig.timeZoneOffset)
				if nodedata:
					uql = ULTIPA.ExportReply(data=nodedata)
					response.data = uql.data
					cb(uql.data,None)
				if edgedata:
					uql = ULTIPA.ExportReply(data=edgedata)
					response.data = uql.data	
					cb(None,uql.data)	

		except Exception as e:
			errorRes = ULTIPA_RESPONSE.Response()
			try:
				message = str(e._state.code) + ' : ' + str(e._state.details)
				print(message)
			except:
				message = 'UNKNOW ERROR'
				print(message)
			errorRes.status = ULTIPA.Status(code=ULTIPA.Code.UNKNOW_ERROR, message=message)
			errorRes.req = ULTIPA.ReturnReq(self.graphSetName, "exportData",
											requestConfig.useHost if requestConfig.useHost else self.host,
											requestConfig.retry,
											False)
			return errorRes

	def uql(self, uql: str,
			requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.UltipaResponse:
		'''
		Execute UQL.

		Args:
			uql: A uql statement

			requestConfig: An object of RequestConfig class

		Returns:
			UltipaResponse

		'''
		request = ultipa_pb2.UqlRequest()
		request.uql = uql
		request.timeout = self.getTimeout(requestConfig.timeoutWithSeconds)
		if requestConfig.threadNum is not None:
			request.thread_num = requestConfig.threadNum
		ultipaRes = ULTIPA_RESPONSE.UltipaResponse()
		uqlLoggerConfig = self.defaultConfig.uqlLoggerConfig
		if requestConfig.graphName == '' and self.defaultConfig.defaultGraph != '':
			requestConfig.graphName = self.defaultConfig.defaultGraph

		if self.defaultConfig.consistency != self.hostManagerControl.consistency:
			self.hostManagerControl.consistency = self.defaultConfig.consistency

		if self.defaultConfig.maxRecvSize != self.hostManagerControl.maxRecvSize:
			self.hostManagerControl.maxRecvSize = self.defaultConfig.maxRecvSize
		onRetry = copy.deepcopy(requestConfig.retry)
		canRetry=True
		while onRetry.current < onRetry.max and canRetry:
			try:
				import pytz
				getTimeZoneOffset(requestConfig, self.defaultConfig)
				timeZone = requestConfig.timeZone if requestConfig.timeZone else self.defaultConfig.timeZone
				timeZoneOffset = requestConfig.timeZoneOffset if requestConfig.timeZoneOffset else self.defaultConfig.timeZoneOffset
				timeZoneOffset = getTimeOffsetSeconds(timeZoneOffset)
				clientInfo = self.getClientInfo(graphSetName=requestConfig.graphName, uql=uql,
												useHost=requestConfig.useHost,
												useMaster=requestConfig.useMaster, timezone=timeZone,
												timeZoneOffset=timeZoneOffset)
				if uqlLoggerConfig:
					uqlLoggerConfig.getlogger().info(
						f'Begin UQL: {uql} graphSetName: {clientInfo.graphSetName} Host: {clientInfo.host}')
				uqlIsExtra = UQLHelper.uqlIsExtra(uql)
				if uqlIsExtra:
					res = clientInfo.Controlsclient.UqlEx(request, metadata=clientInfo.metadata)
				else:
					res = clientInfo.Rpcsclient.Uql(request, metadata=clientInfo.metadata)

				if not requestConfig.stream:
					ultipaRes = FormatType.uqlMergeResponse(res, timeZone, timeZoneOffset)
				else:
					ultipaRes = FormatType.uqlResponse(res, timeZone, timeZoneOffset)

				if self.defaultConfig.responseWithRequestInfo and not requestConfig.stream:
					ultipaRes.req = ULTIPA.ReturnReq(clientInfo.graphSetName, uql, clientInfo.host, onRetry,
													 uqlIsExtra)
				if not isinstance(ultipaRes, types.GeneratorType) and RetryHelp.checkRes(ultipaRes):
					onRetry.current += 1
					continue
				else:
					return ultipaRes

			except Exception as e:
				try:
					message = str(e._state.code) + ' : ' + str(e._state.details)
				except:
					message = str(e)
				ultipaRes.status = ULTIPA.Status(code=ULTIPA.Code.UNKNOW_ERROR, message=message)
				ultipaRes.req = ULTIPA.ReturnReq(requestConfig.graphName, uql,
												 requestConfig.useHost if requestConfig.useHost else self.host,
												 onRetry, False)
				if ultipaRes.status.code not in {ULTIPA.Code.RAFT_REDIRECT, ULTIPA.Code.RAFT_LEADER_NOT_YET_ELECTED, ULTIPA.Code.RAFT_NO_AVAILABLE_FOLLOWERS,ULTIPA.Code.RAFT_NO_AVAILABLE_ALGO_SERVERS}:
					canRetry=False

				else:
					onRetry.current += 1
					if uqlLoggerConfig:
						uqlLoggerConfig.getlogger().info(
							f'Begin Retry [{onRetry.current}]- clientInfo host: {clientInfo.host} graphSetName: {clientInfo.graphSetName}')
					self.hostManagerControl.getHostManger(requestConfig.graphName).raftReady = False					

			

		return ultipaRes
	
	def uqlStream(self,uql:str,stream: UQLResponseStream,requestConfig: RequestConfig = RequestConfig()):
		
		'''
		Execute UQL.

		Args:
			uql: A uql statement

			requestConfig: An object of RequestConfig class

		Returns:
			Stream

		'''
		stream.emit("start", requestConfig)
		request = ultipa_pb2.UqlRequest()
		request.uql = uql
		request.timeout = self.getTimeout(requestConfig.timeoutWithSeconds)
		if requestConfig.threadNum is not None:
			request.thread_num = requestConfig.threadNum
		uqlLoggerConfig = self.defaultConfig.uqlLoggerConfig
		if requestConfig.graphName == '' and self.defaultConfig.defaultGraph != '':
			requestConfig.graphName = self.defaultConfig.defaultGraph

		if self.defaultConfig.consistency != self.hostManagerControl.consistency:
			self.hostManagerControl.consistency = self.defaultConfig.consistency

		if self.defaultConfig.maxRecvSize != self.hostManagerControl.maxRecvSize:
			self.hostManagerControl.maxRecvSize = self.defaultConfig.maxRecvSize
		onRetry = copy.deepcopy(requestConfig.retry)
		while onRetry.current < onRetry.max:
			try:
				import pytz
				getTimeZoneOffset(requestConfig, self.defaultConfig)
				timeZone = requestConfig.timeZone if requestConfig.timeZone else self.defaultConfig.timeZone
				timeZoneOffset = requestConfig.timeZoneOffset if requestConfig.timeZoneOffset else self.defaultConfig.timeZoneOffset
				timeZoneOffset = getTimeOffsetSeconds(timeZoneOffset)
				clientInfo = self.getClientInfo(graphSetName=requestConfig.graphName, uql=uql,
												useHost=requestConfig.useHost,
												useMaster=requestConfig.useMaster, timezone=timeZone,
												timeZoneOffset=timeZoneOffset)
				if uqlLoggerConfig:
					uqlLoggerConfig.getlogger().info(
						f'Begin UQL: {uql} graphSetName: {clientInfo.graphSetName} Host: {clientInfo.host}')
				uqlIsExtra = UQLHelper.uqlIsExtra(uql)
				if uqlIsExtra:
					res = clientInfo.Controlsclient.UqlEx(request, metadata=clientInfo.metadata)
				else:
					res = clientInfo.Rpcsclient.Uql(request, metadata=clientInfo.metadata)

				uql_response = ULTIPA_RESPONSE.Response()
				ultipa_response = ULTIPA_RESPONSE.UltipaResponse()
				for uqlReply in res:
					status = FormatType.status(uqlReply.status)
					uql_response = FormatType.response(uql_response, uqlReply, timeZone, timeZoneOffset)
					ret = ULTIPA.UqlReply(dataBase=uql_response.data)

					if status.code != ULTIPA.Code.SUCCESS:
						ultipa_response.status = uql_response.status
						ultipa_response.req = uql_response.req
						stream.emit("end", requestConfig)
						return 

					ultipa_response.items = ret._aliasMap
					ultipa_response.status = uql_response.status
					ultipa_response.req = uql_response.req
					ultipa_response.statistics = uql_response.statistics
					should_continue=stream.emit("data", ultipa_response, requestConfig)	
					if should_continue==False:
						stream.emit("end", requestConfig)	
						return 

				if not isinstance(ultipa_response, types.GeneratorType) and RetryHelp.checkRes(ultipa_response):
					onRetry.current += 1
					continue
				else:
					stream.emit("end", requestConfig)
					return 

			except Exception as e:
				ultipaRes = ULTIPA_RESPONSE.UltipaResponse()
				onRetry.current += 1
				if uqlLoggerConfig:
					uqlLoggerConfig.getlogger().info(
						f'Begin Retry [{onRetry.current}]- clientInfo host: {clientInfo.host} graphSetName: {clientInfo.graphSetName}')
				self.hostManagerControl.getHostManger(requestConfig.graphName).raftReady = False
				try:
					message = str(e._state.code) + ' : ' + str(e._state.details)
				except:
					message = str(e)
				ultipaRes.status = ULTIPA.Status(code=ULTIPA.Code.UNKNOW_ERROR, message=message)
				ultipaRes.req = ULTIPA.ReturnReq(requestConfig.graphName, uql,
												 requestConfig.useHost if requestConfig.useHost else self.host,
												 onRetry, False)
				print(ultipaRes.status.message)
		stream.emit("end", requestConfig)
		return 

	def uqlSingle(self, uqlMaker: UQLMAKER) -> ULTIPA_RESPONSE.UltipaResponse:
		res = self.uql(uqlMaker.toString(), uqlMaker.commonParams)
		return res


	def UqlListSimple(self, uqlMaker: UQLMAKER, responseKeyFormat: ResponseKeyFormat = None,
					  isSingleOne: bool = True) -> ULTIPA_RESPONSE.UltipaResponse:
		res = self.uqlSingle(uqlMaker)
		if res.status.code != ULTIPA.Code.SUCCESS:
			simplrRes = ULTIPA_RESPONSE.Response(res.status, res.items)
			return simplrRes

		if not isSingleOne:
			retList = []
			for alias in res.aliases:
				item = res.items.get(alias.alias)
				table = item.data
				table_rows = table.rows
				table_rows_dict = convertTableToDict(table_rows, table.headers)
				if responseKeyFormat:
					table_rows_dict = responseKeyFormat.changeKeyValue(table_rows_dict)
				data = convertToListAnyObject(table_rows_dict)
				retList.append(PropertyTable(name=table.name, data=data))
			simplrRes = ULTIPA_RESPONSE.Response(res.status, retList)
			simplrRes.req = res.req
			return simplrRes

		alisFirst = res.aliases[0].alias if len(res.aliases) > 0 else None
		firstItem = res.items.get(alisFirst)
		if firstItem:
			table_rows = firstItem.data.rows
			table_rows_dict = convertTableToDict(table_rows, firstItem.data.headers)
			if responseKeyFormat:
				table_rows_dict = responseKeyFormat.changeKeyValue(table_rows_dict)
			data = convertToListAnyObject(table_rows_dict)
			simplrRes = ULTIPA_RESPONSE.Response(res.status, data)
			simplrRes.req = res.req
			simplrRes.statistics = res.statistics
			return simplrRes
		else:
			return res

	def UqlUpdateSimple(self, uqlMaker: UQLMAKER):
		res = self.uqlSingle(uqlMaker)

		if res.status.code != ULTIPA.Code.SUCCESS:
			return ULTIPA_RESPONSE.Response(res.status, statistics=res.statistics)

		if res.req:
			ret = ULTIPA_RESPONSE.Response(res.status, statistics=res.statistics)
			ret.req = res.req
			return ret
		return ULTIPA_RESPONSE.Response(res.status, statistics=res.statistics)

	def insertNodesBatchBySchema(self, schema: Schema, rows: List[Node],
								 config: InsertConfig) -> ULTIPA_RESPONSE.InsertResponse:
		'''
		Batch insert nodes of a same schema (that already exists in the graphset).

		Args:
			schema: An object of Schema class

			rows: The data rows to be inserted, List[Node]

			config: An object of InsertConfig class

		Returns:
			InsertResponse

		'''

		config.useMaster = True
		if config.graphName == '' and self.defaultConfig.defaultGraph != '':
			config.graphName = self.defaultConfig.defaultGraph

		clientInfo = self.getClientInfo(clientType=ClientType.Update, graphSetName=config.graphName,
										useMaster=config.useMaster)

		nodetable = FormatType.makeEntityNodeTable(schema, rows, getTimeZoneOffset(requestConfig=config,
																				   defaultConfig=self.defaultConfig))

		_nodeTable = ultipa_pb2.EntityTable(schemas=nodetable.schemas, entity_rows=nodetable.nodeRows)
		request = ultipa_pb2.InsertNodesRequest()
		request.silent = config.silent
		request.insert_type = config.insertType
		request.graph_name = config.graphName
		request.node_table.MergeFrom(_nodeTable)
		res = clientInfo.Rpcsclient.InsertNodes(request, metadata=clientInfo.metadata)

		status = FormatType.status(res.status)
		uqlres = ULTIPA_RESPONSE.Response(status=status)
		reTry = RetryHelp.check(self, config, uqlres)
		if reTry.canRetry:
			config.retry = reTry.nextRetry
			return self.insertNodesBatchBySchema(schema, rows, config)

		uRes = ULTIPA_RESPONSE.ResponseBulk()
		uRes.uuids = [i for i in res.uuids]
		errorDict = {}
		for i, data in enumerate(res.ignore_error_code):
			try:
				index = rows[res.ignore_indexes[i]]._getIndex()
			except Exception as e:
				try:
					index = res.ignore_indexes[i]
				except Exception as e:
					index = i
			if index is None:
				try:
					index = res.ignore_indexes[i]
				except Exception as e:
					index = i
			errorDict.update({index: data})
		uRes.errorItem = errorDict
		uqlres.data = uRes
		uqlres.total_time = res.time_cost
		uqlres.engine_time = res.engine_time_cost
		if self.defaultConfig.responseWithRequestInfo:
			uqlres.req = ULTIPA.ReturnReq(config.graphName, "InsertNodesBatchBySchema", clientInfo.host, reTry,
										  False)
		return uqlres

	def insertEdgesBatchBySchema(self, schema: Schema, rows: List[Edge],
								 config: InsertConfig) -> ULTIPA_RESPONSE.InsertResponse:
		'''
		Batch insert edges of a same schema (that already exists in the graphset)

		Args:
			schema: Schema 实例化Schema对象

			rows: The data rows to be inserted, List[Edge]

			config: An object of InsertConfig class

		Returns:
			InsertResponse

		'''
		config.useMaster = True
		if config.graphName == '' and self.defaultConfig.defaultGraph != '':
			config.graphName = self.defaultConfig.defaultGraph

		clientInfo = self.getClientInfo(clientType=ClientType.Update, graphSetName=config.graphName,
										useMaster=config.useMaster)

		edgetable = FormatType.makeEntityEdgeTable(schema=schema, rows=rows,
												   timeZoneOffset=getTimeZoneOffset(requestConfig=config,
																					defaultConfig=self.defaultConfig))

		_edgeTable = ultipa_pb2.EntityTable(schemas=edgetable.schemas, entity_rows=edgetable.edgeRows)
		request = ultipa_pb2.InsertEdgesRequest()
		request.silent = config.silent
		request.insert_type = config.insertType
		request.graph_name = config.graphName
		request.create_node_if_not_exist = config.createNodeIfNotExist
		request.edge_table.MergeFrom(_edgeTable)
		res = clientInfo.Rpcsclient.InsertEdges(request, metadata=clientInfo.metadata)

		status = FormatType.status(res.status)
		uqlres = ULTIPA_RESPONSE.Response(status=status)
		reTry = RetryHelp.check(self, config, uqlres)
		if reTry.canRetry:
			config.retry = reTry.nextRetry
			return self.insertEdgesBatchBySchema(schema, rows, config)

		uRes = ULTIPA_RESPONSE.ResponseBulk()
		uRes.uuids = [i for i in res.uuids]
		errorDict = {}
		for i, data in enumerate(res.ignore_error_code):
			try:
				index = rows[res.ignore_indexes[i]]._getIndex()
			except Exception as e:
				try:
					index = res.ignore_indexes[i]
				except Exception as e:
					index = i
			if index is None:
				try:
					index = res.ignore_indexes[i]
				except Exception as e:
					index = i
			errorDict.update({index: data})
		uRes.errorItem = errorDict
		uqlres.data = uRes
		uqlres.total_time = res.time_cost
		uqlres.engine_time = res.engine_time_cost
		if self.defaultConfig.responseWithRequestInfo:
			uqlres.req = ULTIPA.ReturnReq(config.graphName, "InsertEdgesBatchBySchema", clientInfo.host, reTry,
										  False)
		return uqlres

	def insertNodesBatchAuto(self, nodes: List[Node],
							 config: InsertConfig) -> ULTIPA_RESPONSE.ResponseBatchAutoInsert:
		'''
		Batch insert nodes of different schemas (that will be created if not existent)

		Args:
			nodes: The data rows to be inserted, List[Node]

			config: An object of InsertConfig class

		Returns:
			ResponseBatchAutoInsert

		'''
		Result = {}
		schemaDict = {}
		batches = {}
		schemaRet = self.uql(GetPropertyBySchema.node, config)
		if schemaRet.status.code == ULTIPA.Code.SUCCESS:
			for aliase in schemaRet.aliases:
				if aliase.alias == '_nodeSchema':
					schemaDict = convertTableToDict(schemaRet.alias(aliase.alias).data.rows,
													schemaRet.alias(aliase.alias).data.headers)
			if not schemaDict:
				raise ParameterException(err='Please create Node Schema.')
		else:
			raise ParameterException(err=schemaRet.status.message)
		for index, node in enumerate(nodes):
			node._index = index
			if batches.get(node.schema) is None:
				batches[node.schema] = ULTIPA_REQUEST.Batch()
				find = list(filter(lambda x: x.get('name') == node.schema, schemaDict))
				if find:
					findSchema = find[0]
					propertyList = FormatType.checkProperty(node, json.loads(findSchema.get("properties")))
					reqSchema = ULTIPA_REQUEST.Schema(name=node.schema, properties=propertyList,dbType=DBType.DBNODE)
					batches[node.schema].Schema = reqSchema
				else:
					if node.schema is None:
						raise ParameterException(err=f"Row [{index}]:Please set schema name for node.")
					else:
						raise ParameterException(err=f"Row [{index}]:Node Schema not found {node.schema}.")

			batches.get(node.schema).Nodes.append(node)
		for key in batches:
			batch = batches.get(key)
			Result.update(
				{key: self.insertNodesBatchBySchema(schema=batch.Schema, rows=batch.Nodes, config=config)})

		newStatusMsg = ""
		newCode = None
		for i, key in enumerate(Result):
			ret = Result.get(key)
			newStatusMsg += f"{key}:{ret.status.message} "
			if ret.status.code != ULTIPA.Code.SUCCESS and not newCode:
				newCode = ret.status.code
		if newCode is None:
			newCode = ULTIPA.Code.SUCCESS
		status = ULTIPA_RESPONSE.Status(newCode, newStatusMsg)
		newResponse = ULTIPA_RESPONSE.ResponseBatchAutoInsert(status=status)
		newResponse.data = Result
		return newResponse

	def insertEdgesBatchAuto(self, edges: List[Edge],
							 config: InsertConfig) -> ULTIPA_RESPONSE.ResponseBatchAutoInsert:
		'''
		Batch insert edges of different schemas (that will be created if not existent)

		Args:
			edges: The data rows to be inserted, List[Edge]

			config: An object of InsertConfig class

		Returns:
			ResponseBatchAutoInsert

		'''
		Result = {}
		schemaDict = []
		batches = {}
		schemaRet = self.uql(GetPropertyBySchema.edge, config)
		if schemaRet.status.code == ULTIPA.Code.SUCCESS:
			for aliase in schemaRet.aliases:
				if aliase.alias == '_edgeSchema':
					schemaDict = convertTableToDict(schemaRet.alias(aliase.alias).data.rows,
													schemaRet.alias(aliase.alias).data.headers)
			if not schemaDict:
				raise ParameterException(err='Please create Edge Schema.')
		else:
			raise ParameterException(err=schemaRet.status.message)
		for index, edge in enumerate(edges):
			edge._index = index
			if batches.get(edge.schema) == None:
				batches[edge.schema] = ULTIPA_REQUEST.Batch()
				find = list(filter(lambda x: x.get('name') == edge.schema, schemaDict))
				if find:
					findSchema = find[0]
					propertyList = FormatType.checkProperty(edge, json.loads(findSchema.get("properties")))
					reqSchema = ULTIPA_REQUEST.Schema(name=edge.schema, properties=propertyList,dbType=DBType.DBEDGE)
					batches[edge.schema].Schema = reqSchema
				else:
					if edge.schema is None:
						raise ParameterException(err=f"Row [{index}]:Please set schema name for edge.")
					else:
						raise ParameterException(err=f"Row [{index}]:Edge Schema not found {edge.schema}.")
			batches.get(edge.schema).Edges.append(edge)
		for key in batches:
			batch = batches.get(key)
			Result.update(
				{key: self.insertEdgesBatchBySchema(schema=batch.Schema, rows=batch.Edges, config=config)})

		newStatusMsg = ""
		newCode = None
		for i, key in enumerate(Result):
			ret = Result.get(key)
			newStatusMsg += f"{key}:{ret.status.message} "
			if ret.status.code != ULTIPA.Code.SUCCESS and not newCode:
				newCode = ret.status.code
		if newCode is None:
			newCode = ULTIPA.Code.SUCCESS
		status = ULTIPA_RESPONSE.Status(newCode, newStatusMsg)
		newResponse = ULTIPA_RESPONSE.ResponseBatchAutoInsert(status=status)
		newResponse.data = Result
		return newResponse

	def _InsertByCSV(self, csvPath: str, type: DBType, config: InsertConfig,
					 schemaName: str = None) -> ULTIPA_RESPONSE.ResponseBatchAutoInsert:
		rows = []
		propertyType = []
		properties = []
		types = []
		with open(csvPath, "r", encoding="utf-8-sig") as csvfile:
			reader = csv.reader(csvfile)
			for i, line in enumerate(reader):
				if i == 0:
					for i, property in enumerate(line):
						k1, k2 = property.split(":")
						propertyType.append({k1: k2})
						types.append({"index": i, "type": k2})
						properties.append(k1)
					continue
				for i in types:
					if i.get("type") in ["int", "int32", "int64","uint32","uint64"]:
						if line[i.get("index")] == "":
							line[i.get("index")] = 0
							continue
						line[i.get("index")] = int(line[i.get("index")])
					if i.get("type") in ["float", "double"]:
						if line[i.get("index")] == "":
							line[i.get("index")] = 0.0
							continue
						line[i.get("index")] = float(line[i.get("index")])
				line = dict(zip(properties, line))
				if i == 0:
					print(line.keys())
				if type == DBType.DBNODE:
					if line.get("_uuid"):
						uuid = line.get("_uuid")
						line.__delitem__("_uuid")
						if line.get("_id"):
							line.__delitem__("_id")
						rows.append(ULTIPA.Node(line, schema_name=schemaName, uuid=int(uuid)))
					elif line.get("_id"):
						id = line.get("_id")
						line.__delitem__("_id")
						rows.append(ULTIPA.Node(line, schema_name=schemaName, id=id))
					else:
						rows.append(ULTIPA.Node(line, schema_name=schemaName))

				elif type == DBType.DBEDGE:
					if line.get("_from_uuid") and line.get("_to_uuid"):
						from_uuid = line.get("_from_uuid")
						line.__delitem__("_from_uuid")
						to_uuid = line.get("_to_uuid")
						line.__delitem__("_to_uuid")
						if line.get("_id"):
							line.__delitem__("_id")
						if line.get("_uuid"):
							line.__delitem__("_uuid")
						if line.get("_from"):
							line.__delitem__("_from")
						if line.get("_to"):
							line.__delitem__("_to")
						rows.append(
							ULTIPA.Edge(line, schema_name=schemaName, from_uuid=int(from_uuid),
										to_uuid=int(to_uuid)))
					elif line.get("_from") and line.get("_to"):
						from_id = line.get("_from")
						line.__delitem__("_from")
						to_id = line.get("_to")
						line.__delitem__("_to")
						if line.get("_id"):
							line.__delitem__("_id")
						if line.get("_uuid"):
							line.__delitem__("_uuid")
						if line.get("_from_uuid"):
							line.__delitem__("_from_uuid")
						if line.get("_to_uuid"):
							line.__delitem__("_to_uuid")
						rows.append(ULTIPA.Edge(line, schema_name=schemaName, from_id=from_id, to_id=to_id))

		if type == DBType.DBNODE:
			return self.insertNodesBatchAuto(rows, config)
		else:
			return self.insertEdgesBatchAuto(rows, config)


# -*- coding: utf-8 -*-
# @Time    : 2023/1/16 11:06 AM
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : propertyUtils.py
from ultipa.structs.PropertyType import PropertyTypeStr
from ultipa.types import ULTIPA
from typing import List


def isBasePropertyType(type: PropertyTypeStr):
	'''
	Judge whether a data type is a base property type (not a list).

	Args:
		type:

	Returns:

	'''
	if type in [PropertyTypeStr.PROPERTY_STRING,
				PropertyTypeStr.PROPERTY_INT,
				PropertyTypeStr.PROPERTY_INT64,
				PropertyTypeStr.PROPERTY_UINT32,
				PropertyTypeStr.PROPERTY_UINT64,
				PropertyTypeStr.PROPERTY_FLOAT,
				PropertyTypeStr.PROPERTY_DOUBLE,
				PropertyTypeStr.PROPERTY_DATETIME,
				PropertyTypeStr.PROPERTY_TIMESTAMP,
				PropertyTypeStr.PROPERTY_TEXT]:
		return True
	return False


def getPropertyTypesDesc(type: PropertyTypeStr, subTypes: List[PropertyTypeStr]):
	'''
	Generate the format string a list type corresponds to.

	Args:
		type:

		subTypes:

	Returns:
		str[]
	'''
	if type == PropertyTypeStr.PROPERTY_LIST:
		subType = subTypes[0]
		if isBasePropertyType(subType):
			return f"{subType}[]"
	if type == PropertyTypeStr.PROPERTY_SET:
		subType = subTypes[0]
		if isBasePropertyType(subType):
			return f"set({subType})"
	return type



# -*- coding: utf-8 -*-
# @Time    : 2023/8/4 16:40
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : PropertyType.py
from ultipa.proto import ultipa_pb2

class PropertyTypeStr:
	'''
	    Data class for property type mapping to string.
	'''
	PROPERTY_INT = 'int32'
	PROPERTY_STRING = 'string'
	PROPERTY_FLOAT = 'float'
	PROPERTY_DOUBLE = 'double'
	PROPERTY_UINT32 = 'uint32'
	PROPERTY_INT64 = 'int64'
	PROPERTY_UINT64 = 'uint64'
	PROPERTY_DATETIME = 'datetime'
	PROPERTY_TIMESTAMP = 'timestamp'
	PROPERTY_TEXT = 'text'
	PROPERTY_UNSET = "unset"
	PROPERTY_POINT = "point"
	PROPERTY_LIST = "list"
	PROPERTY_SET = "set"
	PROPERTY_MAP = "map"
	PROPERTY_BLOB= "blob"
	@staticmethod
	def PROPERTY_DECIMAL(accuracy,scale):
		return f"decimal({accuracy},{scale})"

class PropertyType:
	'''
	    Data class for property type mapping to gRPC.
	'''
	PROPERTY_UNSET = ultipa_pb2.UNSET
	PROPERTY_INT32 = ultipa_pb2.INT32
	PROPERTY_STRING = ultipa_pb2.STRING
	PROPERTY_FLOAT = ultipa_pb2.FLOAT
	PROPERTY_DOUBLE = ultipa_pb2.DOUBLE
	PROPERTY_UINT32 = ultipa_pb2.UINT32
	PROPERTY_INT64 = ultipa_pb2.INT64
	PROPERTY_UINT64 = ultipa_pb2.UINT64
	PROPERTY_DATETIME = ultipa_pb2.DATETIME
	PROPERTY_TIMESTAMP = ultipa_pb2.TIMESTAMP
	PROPERTY_TEXT = ultipa_pb2.TEXT
	PROPERTY_BLOB = ultipa_pb2.BLOB
	PROPERTY_POINT = ultipa_pb2.POINT
	PROPERTY_DECIMAL = ultipa_pb2.DECIMAL
	PROPERTY_LIST = ultipa_pb2.LIST
	PROPERTY_SET = ultipa_pb2.SET
	PROPERTY_MAP = ultipa_pb2.MAP
	PROPERTY_NULL = ultipa_pb2.NULL_
	PROPERTY_UUID = -1
	PROPERTY_ID = -2
	PROPERTY_FROM = -3
	PROPERTY_FROM_UUID = -4
	PROPERTY_TO = -5
	PROPERTY_TO_UUID = -6
	PROPERTY_IGNORE = -7


from ultipa.operate.base_extra import BaseExtra
from ultipa.utils import UQLMAKER, CommandList
from ultipa.types import ULTIPA_REQUEST, ULTIPA_RESPONSE
from ultipa.utils.errors import ParameterException
from ultipa.utils.format import FormatType
from ultipa.utils.ResposeFormat import ResponseKeyFormat
from ultipa.configuration.RequestConfig import RequestConfig
from ultipa.proto import ultipa_pb2
from ultipa.connection.clientType import ClientType
from ultipa.types import ULTIPA, ULTIPA_REQUEST, ULTIPA_RESPONSE
from ultipa.utils.convert import convertToUser

JSONSTRING_KEYS = ["graphPrivileges", "systemPrivileges", "policies", "policy", "privilege"]
formatdata = ['graph_privileges']


class UserExtra(BaseExtra):
	'''
		Processing class that defines settings for user related operations.
	'''

	def _GRPATH_PRIVILEGES_DATA_FORMAT(self, obj):
		if isinstance(obj.get('graph_privileges'), list):
			resr = FormatType.graphPrivileges(obj.get('graph_privileges'))
			return resr
		else:
			return '[]'

	def showUser(self,
				 requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.ResponseListUser:
		'''
		Show user list.

		Args:
			requestConfig: An object of RequestConfig class

		Returns:
			ResponseListUser

		'''

		uqlMaker = UQLMAKER(command=CommandList.showUser, commonParams=requestConfig)
		res = self.UqlListSimple(uqlMaker=uqlMaker,
								 responseKeyFormat=ResponseKeyFormat(jsonKeys=JSONSTRING_KEYS, dataFormat=formatdata))
		if len(res.data)>0:
			res.data=convertToUser(res)
		return res

	def getUser(self, userName:str,
				requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.ResponseUser:
		'''
		Get a designated user.

		Args:
			username: The name of user

			requestConfig: An object of RequestConfig class

		Returns:
			ResponseUser

		'''

		uqlMaker = UQLMAKER(command=CommandList.getUser, commonParams=requestConfig)
		uqlMaker.setCommandParams(userName)
		res = self.UqlListSimple(uqlMaker=uqlMaker,
								 responseKeyFormat=ResponseKeyFormat(jsonKeys=JSONSTRING_KEYS, dataFormat=formatdata))
		# if isinstance(res.data, list) and len(res.data) > 0:
		# 	res.data = res.data[0]
		if res.status==ULTIPA.Code.SUCCESS:
			res.data=convertToUser(res)
		return res

	def getSelfInfo(self,
					requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.UltipaResponse:
		'''
		Get the current user.

		Args:
			requestConfig: An object of RequestConfig class

		Returns:
			UltipaResponse

		'''
		uqlMaker = UQLMAKER(command=CommandList.getSelfInfo, commonParams=requestConfig)
		res = self.UqlListSimple(uqlMaker=uqlMaker,
								 responseKeyFormat=ResponseKeyFormat(jsonKeys=JSONSTRING_KEYS, dataFormat=formatdata))

		return res

	def createUser(self, request: ULTIPA_REQUEST.CreateUser,
				   requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.UltipaResponse:
		'''
		Create a user.

		Args:
			request: An object of CreateUser class

			requestConfig: An object of RequestConfig class

		Returns:
			UltipaResponse

		'''

		uqlMaker = UQLMAKER(command=CommandList.createUser, commonParams=requestConfig)
		params = []
		if request.username:
			params.append(request.username)
		else:
			raise ParameterException(err='username is a required parameter')

		if request.password:
			params.append(request.password)
		else:
			raise ParameterException(err='password is a required parameter')

		if request.graph_privileges:
			params.append(request.graph_privileges)
		else:
			params.append({})

		if request.system_privileges:
			params.append(request.system_privileges)
		else:
			params.append([])

		if request.policies:
			params.append(request.policies)
		else:
			params.append([])	

		if request.property_privileges:
			params.append(request.property_privileges)
		else:
			params.append({})


		uqlMaker.setCommandParams(params)
		return self.uqlSingle(uqlMaker=uqlMaker)

	def dropUser(self, userName:str,
				 requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.UltipaResponse:
		'''
		Drop a user.

		Args:
			userName: The name of user

			requestConfig: An object of RequestConfig class

		Returns:
			UltipaResponse

		'''

		uqlMaker = UQLMAKER(command=CommandList.dropUser, commonParams=requestConfig)
		uqlMaker.setCommandParams(userName)
		return self.uqlSingle(uqlMaker=uqlMaker)

	def alterUser(self, request: ULTIPA_REQUEST.AlterUser,
				  requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.UltipaResponse:
		'''
		Alter a user.

		Args:
			request: An object of AlterUser class

			requestConfig: An object of RequestConfig class

		Returns:
			UltipaResponse

		'''

		uqlMaker = UQLMAKER(command=CommandList.alterUser, commonParams=requestConfig)
		if request.username:
			uqlMaker.setCommandParams(request.username)
		else:
			raise ParameterException(err='username is a required parameter')

		paramsDict = {}
		if request.password:
			paramsDict.setdefault('password', request.password)

		if request.graph_privileges:
			paramsDict.setdefault('graph_privileges', request.graph_privileges)

		if request.system_privileges:
			paramsDict.setdefault('system_privileges', request.system_privileges)

		if request.policies:
			paramsDict.setdefault('policies', request.policies)

		if request.property_privileges:
			paramsDict.setdefault('property_privileges', request.property_privileges)				

		uqlMaker.addParam('set', paramsDict)
		return self.uqlSingle(uqlMaker=uqlMaker)

	def getUserSetting(self,request:ULTIPA_REQUEST.getUserSetting,
					 requestconfig:RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.UltipaResponse:
		
		'''
		Get User Settings .

		Args:
			request: An object of getUserSetting Class

			requestConfig: An object of RequestConfig class

		Returns:
			UltipaResponse

		'''
		
		try:
			request=ultipa_pb2.UserSettingRequest(user_name=request.username,type=request.type,
										 opt=ultipa_pb2.UserSettingRequest.OPTION.OPT_GET)
			clientInfo = self.getClientInfo(graphSetName=requestconfig.graphName, useMaster=requestconfig.useMaster, clientType=ClientType.Leader,
											isGlobal=True,timezone=requestconfig.timeZone, timeZoneOffset=requestconfig.timeZoneOffset)	
			res=clientInfo.Controlsclient.UserSetting(request,metadata=clientInfo.metadata)
			status = FormatType.status(res.status)
			data=res.data
			response = ULTIPA_RESPONSE.Response(status=status,data=data)
			return response
		
		except Exception as e:
			errorRes = ULTIPA_RESPONSE.Response()
			try:
				message = str(e._state.code) + ' : ' + str(e._state.details)
			except:
				message = 'UNKNOWN ERROR'
			errorRes.status = ULTIPA.Status(code=ULTIPA.Code.UNKNOW_ERROR, message=message)	
			return errorRes	
		
	def setUserSetting(self,request:ULTIPA_REQUEST.setUserSetting,
					 requestconfig:RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.UltipaResponse:
		
		'''
		set User Settings .

		Args:
			request: An object of setUserSetting Class

			requestConfig: An object of RequestConfig class

		Returns:
			UltipaResponse

		'''
		try:
			request=ultipa_pb2.UserSettingRequest(user_name=request.username,type=request.type,
										 opt=ultipa_pb2.UserSettingRequest.OPTION.OPT_SET,data=request.data)
			clientInfo = self.getClientInfo(graphSetName=requestconfig.graphName, useMaster=requestconfig.useMaster,
								    clientType=ClientType.Leader,isGlobal=True,
									timezone=requestconfig.timeZone, timeZoneOffset=requestconfig.timeZoneOffset)	
			res=clientInfo.Controlsclient.UserSetting(request,metadata=clientInfo.metadata)
			status = FormatType.status(res.status)
			response = ULTIPA_RESPONSE.Response(status=status)
			return response



		except Exception as e:
			errorRes = ULTIPA_RESPONSE.Response()
			try:
				message = str(e._state.code) + ' : ' + str(e._state.details)
			except:
				message = 'UNKNOWN ERROR'
			errorRes.status = ULTIPA.Status(code=ULTIPA.Code.UNKNOW_ERROR, message=message)	
			return errorRes	



			

		




# -*- coding: utf-8 -*-
import datetime

from ultipa.types import ULTIPA
from ultipa.utils.checkStrTime import is_valid_date


class TypeCheck:
	'''
	Check the data type.
	'''

	@staticmethod
	def checkProperty(type, value):
		if type == ULTIPA.PropertyType.PROPERTY_UINT32:
			if isinstance(value, int) or value is None:
				return True
			else:
				return "%s row [%s] error: failed to serialize value of property %s [uint32],value=%s"

		if type == ULTIPA.PropertyType.PROPERTY_UINT64:
			if isinstance(value, int) or value is None:
				return True
			else:
				return "%s row [%s] error: failed to serialize value of property %s [uint64],value=%s"

		if type == ULTIPA.PropertyType.PROPERTY_INT32:
			if isinstance(value, int) or value is None:
				return True
			else:
				return "%s row [%s] error: failed to serialize value of property %s [int32],value=%s"

		if type == ULTIPA.PropertyType.PROPERTY_INT64:
			if isinstance(value, int) or value is None:
				return True
			else:
				return "%s row [%s] error: failed to serialize value of property %s [int64],value=%s"

		if type in [ULTIPA.PropertyType.PROPERTY_STRING, ULTIPA.PropertyType.PROPERTY_TO,
					ULTIPA.PropertyType.PROPERTY_FROM, ULTIPA.PropertyType.PROPERTY_ID,ULTIPA.PropertyType.PROPERTY_TEXT
					]:
			if isinstance(value, str) or value is None:
				return True
			else:
				return "%s row [%s] error: failed to serialize value of property %s [string],value=%s"

		if type in [ULTIPA.PropertyType.PROPERTY_UUID, ULTIPA.PropertyType.PROPERTY_FROM_UUID, ULTIPA.PropertyType.PROPERTY_TO_UUID]:
			if isinstance(value, int) or value is None:
				return True
			else:
				return "%s row [%s] error: failed to serialize value of property %s [int64],value=%s"

		if type in [ULTIPA.PropertyType.PROPERTY_FLOAT]:
			if isinstance(value, float) or value is None:
				return True
			else:
				return "%s row [%s] error: failed to serialize value of property %s [float],value=%s"

		if type in [ULTIPA.PropertyType.PROPERTY_DOUBLE]:
			if isinstance(value, float) or value is None:
				return True
			else:
				return "%s row [%s] error: failed to serialize value of property %s [double],value=%s"

		if type == ULTIPA.PropertyType.PROPERTY_DATETIME:
			if is_valid_date(value) != False or isinstance(value, datetime.datetime) or value is None:
				return True
			else:
				return "%s row [%s] error: failed to serialize value of property %s [datetime],value=%s"

		if type == ULTIPA.PropertyType.PROPERTY_TIMESTAMP:
			if is_valid_date(value) != False or isinstance(value, datetime.datetime) or isinstance(value, int) or value is None:
				return True
			else:
				return "%s row [%s] error: failed to serialize value of property %s [timestamp],value=%s"

		return True

if __name__ == '__main__':
	ret = TypeCheck.checkProperty(1,"2019-12-12 15:59:59")
	print(ret)



# -*- coding: utf-8 -*-
# @Time    : 2024/05/14 11:47
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : Index.py
from typing import List

from ultipa.structs import DBType
from ultipa.structs.BaseModel import BaseModel


class Index(BaseModel):
	'''
	    Data class for Index.
	'''

	def __init__(self, name: str, properties:str=None,schema:str=None, status:str=None,dbType:DBType=None,size:str=''):
		self.name = name
		self.properties = properties
		self.schema = schema
		self.status = status
		self.size = size
		self.DBType = dbType

	


import hashlib
import os

from ultipa.configuration.RequestConfig import RequestConfig
from ultipa.operate.base_extra import BaseExtra
from ultipa.proto import ultipa_pb2
from ultipa.structs.Algo import ALGO_RESULT
from ultipa.types import ULTIPA_REQUEST, ULTIPA, ULTIPA_RESPONSE
from ultipa.utils import UQLMAKER, CommandList, errors
from ultipa.utils.fileSize import read_in_chunks
from ultipa.utils.format import FormatType
from ultipa.utils.ResposeFormat import ResponseKeyFormat
from ultipa.utils.convert import convertToAlgo


class AlgoExtra(BaseExtra):
	'''
		Processing class that defines settings for algorithm related operations.

	'''
	JSONSTRING_KEYS = ["param"]

	def showAlgo(self,
				 config: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.ResponseListAlgo:
		'''
		Query for algorithm list.

		Args:
			config: An object of the RequestConfig class

		Returns:
			ResponseListAlgo

		'''
		uqlMaker = UQLMAKER(command=CommandList.showAlgo, commonParams=config)
		res = self.UqlListSimple(uqlMaker=uqlMaker, responseKeyFormat=ResponseKeyFormat(jsonKeys=self.JSONSTRING_KEYS))
		if res.status.code == ULTIPA.Code.SUCCESS:
			res.data=convertToAlgo(res)
			for algo in res.data:
				try:
					result_opt = int(algo.result_opt) if algo.result_opt is not None else 0
				except Exception as e:
					raise errors.ParameterException(e)
				result_opt_obj = ULTIPA_RESPONSE.AlgoResultOpt()
				result_opt_obj.can_realtime = True if result_opt & ALGO_RESULT.WRITE_TO_CLIENT else False
				result_opt_obj.can_visualization = True if result_opt & ALGO_RESULT.WRITE_TO_VISUALIZATION else False
				result_opt_obj.can_write_back = True if result_opt & (
						ALGO_RESULT.WRITE_TO_DB | ALGO_RESULT.WRITE_TO_FILE) else False
				algo.__setattr__("result_opt", result_opt_obj)
		return res
	
	def showExta(self,config: RequestConfig = RequestConfig())->ULTIPA_RESPONSE.ResponeListExta:
		'''
		Query for algorithm list.

		Args:
			config: An object of the RequestConfig class

		Returns:
			ResponseListAlgo

		'''
		uqlMaker = UQLMAKER(command=CommandList.showExta, commonParams=config)
		res = self.UqlListSimple(uqlMaker)
		return res
		

	def __make_message(self, filename, md5, chunk):
		return ultipa_pb2.InstallAlgoRequest(
			file_name=filename, md5=md5, chunk=chunk
		)

	def __generate_messages(self, request: ULTIPA_REQUEST.InstallAlgo):
		messages = []
		file_object = open(request.soPath, 'rb')
		somd5 = hashlib.md5(file_object.read()).hexdigest()
		file_object.close()

		for chunk in read_in_chunks(request.soPath):
			filename = os.path.basename(request.soPath)
			messages.append(self.__make_message(filename, somd5, chunk))

		file_object = open(request.configPath, 'rb')
		configmd5 = hashlib.md5(file_object.read()).hexdigest()
		file_object.close()
		for chunk in read_in_chunks(request.configPath):
			filename = os.path.basename(request.configPath)
			messages.append(self.__make_message(filename, configmd5, chunk))
		for msg in messages:
			yield msg

	def installAlgo(self, soFilePath: str, infoFilePath: str,
					config: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.Response:
		'''
		Install an Ultipa standard algorithm.

		Args:
			soFilePath: The directory of the algorithm package (.so)

			infoFilePath: The directory of the algorithm configuration (.yml)

			config: An object of the RequestConfig class

		Returns:
			Response

		'''
		request = ULTIPA_REQUEST.InstallAlgo(infoFilePath, soFilePath)
		config.useMaster = True

		clientInfo = self.getClientInfo(graphSetName=config.graphName, useMaster=config.useMaster,
										isGlobal=True)
		response = ULTIPA_RESPONSE.Response()
		try:
			if os.path.exists(request.soPath) and os.path.exists(request.configPath):
				installRet = clientInfo.Controlsclient.InstallAlgo(self.__generate_messages(request),
																   metadata=clientInfo.metadata)
				status = FormatType.status(installRet.status)
				response.status = status
		except Exception as e:
			try:
				message = str(e._state.code) + ' : ' + str(e._state.details)
			except:
				message = str(e)

			response.status = ULTIPA.Status(code=ULTIPA.Code.UNKNOW_ERROR, message=message)

		if self.defaultConfig.responseWithRequestInfo:
			response.req = ULTIPA.ReturnReq(self.graphSetName, "InstallAlgo",
											config.useHost if config.useHost else self.host,
											config.retry,
											False)
		return response

	def uninstallAlgo(self, algoName: str,
					  config: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.Response:
		'''
		Uninstall an Ultipa standard algorithm.

		Args:
			algoName: The name of algorithm

			config: An object of the RequestConfig class

		Returns:
			Response

		'''
		config.useMaster = True
		clientInfo = self.getClientInfo(graphSetName=config.graphName, useMaster=config.useMaster,
										isGlobal=True)
		arequest = ultipa_pb2.UninstallAlgoRequest(algo_name=algoName)
		installRet = clientInfo.Controlsclient.UninstallAlgo(arequest, metadata=clientInfo.metadata)
		status = FormatType.status(installRet.status)
		response = ULTIPA_RESPONSE.Response(status=status)
		if self.defaultConfig.responseWithRequestInfo:
			response.req = ULTIPA.ReturnReq(self.graphSetName, "UninstallAlgo",
											config.useHost if config.useHost else self.host,
											config.retry,
											False)
		return response

	def installExta(self, extaFilePath: str, extaInfoFilePath: str,
						config: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.Response:
		'''
		Install an EXTA algorithm (via gRPC port).

		Args:
			extaFilePath: The directory of the exta package (.so)

			extaInfoFilePath: The directory of the exta configuration (.yml)

			config: An object of the RequestConfig class

		Returns:
			Response

		'''

		request = ULTIPA_REQUEST.InstallAlgo(extaInfoFilePath, extaFilePath)
		config.useMaster = True
		clientInfo = self.getClientInfo(graphSetName=config.graphName, useMaster=config.useMaster)
		response = ULTIPA_RESPONSE.Response()
		try:
			if os.path.exists(request.soPath) and os.path.exists(request.configPath):
				installRet = clientInfo.Controlsclient.InstallExta(self.__generate_messages(request),
																   metadata=clientInfo.metadata)
				status = FormatType.status(installRet.status)
				response.status = status
		except Exception as e:
			try:
				message = str(e._state.code) + ' : ' + str(e._state.details)
			except:
				message = str(e)

			response.status = ULTIPA.Status(code=ULTIPA.Code.UNKNOW_ERROR, message=message)

		if self.defaultConfig.responseWithRequestInfo:
			response.req = ULTIPA.ReturnReq(self.graphSetName, "InstallExta",
											config.useHost if config.useHost else self.host,
											config.retry,
											False)
		return response

	def uninstallExta(self, extaName: str,
						  requestConfig: RequestConfig = RequestConfig()) -> ULTIPA_RESPONSE.Response:
		'''
		Uninstall an EXTA (via gRPC port).

		Args:
			extaName: The name of exta

			requestConfig: An object of the RequestConfig class

		Returns:
			Response

		'''
		requestConfig.useMaster = True
		clientInfo = self.getClientInfo(graphSetName=requestConfig.graphName, useMaster=requestConfig.useMaster)
		arequest = ultipa_pb2.UninstallExtaRequest(exta_name=extaName)
		installRet = clientInfo.Controlsclient.UninstallExta(arequest, metadata=clientInfo.metadata)
		status = FormatType.status(installRet.status)
		response = ULTIPA_RESPONSE.Response(status=status)
		if self.defaultConfig.responseWithRequestInfo:
			response.req = ULTIPA.ReturnReq(self.graphSetName, "UninstallExta",
											requestConfig.useHost if requestConfig.useHost else self.host,
											requestConfig.retry,
											False)
		return response



# -*- coding: utf-8 -*-
# @Time    : 2023/8/4 16:21
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : EntityRow.py
from typing import Dict


class EntityRow:
	'''
	    Data class for data rows (nodes or edges) to be inserted.
	'''
	_index = None
	def __init__(self, values: Dict, schema: str = None, id: str = None, from_id: str = None, to_id: str = None,
				 uuid: int = None, from_uuid: int = None, to_uuid: int = None, **kwargs):
		self.uuid = uuid
		self.id = id
		self.from_uuid = from_uuid
		self.to_uuid = to_uuid
		self.from_id = from_id
		self.to_id = to_id
		if schema is None:
			if kwargs.get("schema_name") is not None:
				self.schema = kwargs.get("schema_name")
			else:
				self.schema = None
		else:
			self.schema = schema
		self.values = values

	def _getIndex(self):
		return self._index


# -*- coding: utf-8 -*-
# @Time    : 2023/8/1 14:46
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : Property.py
from typing import List

from ultipa.structs.BaseModel import BaseModel
from ultipa.structs.PropertyType import PropertyTypeStr, PropertyType

class Property:
	'''
	    Data class for property.
	'''
	name: str
	description: str
	lte: bool
	schema: str
	type: PropertyTypeStr
	subTypes: List[PropertyTypeStr]

	def __init__(self, name: str,
				 type: PropertyTypeStr = PropertyTypeStr.PROPERTY_STRING,
				 description: str = '', subTypes: List[PropertyTypeStr] = None,lte: bool = None,schema: str=None):
		self.type = type
		self.subTypes = subTypes
		self.description = description
		self.name = name
		self.lte = lte
		self.schema = schema

	_PropertyMap = {
		"string": PropertyType.PROPERTY_STRING,
		"int32": PropertyType.PROPERTY_INT32,
		"int64": PropertyType.PROPERTY_INT64,
		"uint32": PropertyType.PROPERTY_UINT32,
		"uint64": PropertyType.PROPERTY_UINT64,
		"float": PropertyType.PROPERTY_FLOAT,
		"double": PropertyType.PROPERTY_DOUBLE,
		"datetime": PropertyType.PROPERTY_DATETIME,
		"timestamp": PropertyType.PROPERTY_TIMESTAMP,
		"text": PropertyType.PROPERTY_TEXT,
		"blob": PropertyType.PROPERTY_BLOB,
		"_id": PropertyType.PROPERTY_ID,
		"_uuid": PropertyType.PROPERTY_UUID,
		"_from": PropertyType.PROPERTY_FROM,
		"_to": PropertyType.PROPERTY_TO,
		"_from_uuid": PropertyType.PROPERTY_FROM_UUID,
		"_to_uuid": PropertyType.PROPERTY_TO_UUID,
		"_ignore": PropertyType.PROPERTY_IGNORE,
		"unset": PropertyType.PROPERTY_UNSET,
		"point": PropertyType.PROPERTY_POINT,
		"decimal": PropertyType.PROPERTY_DECIMAL,
		"list": PropertyType.PROPERTY_LIST,
		"set": PropertyType.PROPERTY_SET,
		"map": PropertyType.PROPERTY_MAP,
		"null": PropertyType.PROPERTY_NULL,
	}

	_PropertyReverseMap = {
		PropertyType.PROPERTY_STRING: "string",
		PropertyType.PROPERTY_INT32: "int32",
		PropertyType.PROPERTY_INT64: "int64",
		PropertyType.PROPERTY_UINT32: "uint32",
		PropertyType.PROPERTY_UINT64: "uint64",
		PropertyType.PROPERTY_FLOAT: "float",
		PropertyType.PROPERTY_DOUBLE: "double",
		PropertyType.PROPERTY_DATETIME: "datetime",
		PropertyType.PROPERTY_TIMESTAMP: "timestamp",
		PropertyType.PROPERTY_TEXT: "text",
		PropertyType.PROPERTY_BLOB: "blob",
		PropertyType.PROPERTY_ID: "_id",
		PropertyType.PROPERTY_UUID: "_uuid",
		PropertyType.PROPERTY_FROM: "_from",
		PropertyType.PROPERTY_TO: "_to",
		PropertyType.PROPERTY_FROM_UUID: "_from_uuid",
		PropertyType.PROPERTY_TO_UUID: "_to_uuid",
		PropertyType.PROPERTY_IGNORE: "_ignore",
		PropertyType.PROPERTY_UNSET: "unset",
		PropertyType.PROPERTY_POINT: "point",
		PropertyType.PROPERTY_DECIMAL: "decimal",
		PropertyType.PROPERTY_LIST: "list",
		PropertyType.PROPERTY_SET: "set",
		PropertyType.PROPERTY_MAP: "map",
		PropertyType.PROPERTY_NULL: "null",
	}

	def setSubTypesbyType(self, type: str):
		if "string" in type:
			self.subTypes = [PropertyType.PROPERTY_STRING]

		if "int32" in type:
			self.subTypes = [PropertyType.PROPERTY_INT32]

		if "uint32" in type:
			self.subTypes = [PropertyType.PROPERTY_UINT32]

		if "int64" in type:
			self.subTypes = [PropertyType.PROPERTY_INT64]

		if "uint64" in type:
			self.subTypes = [PropertyType.PROPERTY_UINT64]

		if "float" in type:
			self.subTypes = [PropertyType.PROPERTY_FLOAT]

		if "double" in type:
			self.subTypes = [PropertyType.PROPERTY_DOUBLE]

		if "datetime" in type:
			self.subTypes = [PropertyType.PROPERTY_DATETIME]

		if "timestamp" in type:
			self.subTypes = [PropertyType.PROPERTY_TIMESTAMP]

		if "text" in type:
			self.subTypes = [PropertyType.PROPERTY_TEXT]

	def isIdType(self) -> bool:
		idTypes = [
			PropertyType.PROPERTY_ID,
			PropertyType.PROPERTY_TO,
			PropertyType.PROPERTY_UUID,
			PropertyType.PROPERTY_FROM,
			PropertyType.PROPERTY_FROM_UUID,
			PropertyType.PROPERTY_TO_UUID,
		]
		return self.type in idTypes

	def isIgnore(self):
		return self.type == PropertyType.PROPERTY_IGNORE

	def setTypeStr(self, value):
		self.type = self.getStringByPropertyType(value)

	def setTypeInt(self, value):
		self.type = self.getPropertyTypeByString(value)

	def getStringType(self):
		return self.getStringByPropertyType(self.type)

	def getPropertyTypeByString(self, v):
		if not self._PropertyMap.get(v):
			if "[" in v:
				self.setSubTypesbyType(v)
				return PropertyType.PROPERTY_LIST

		return self._PropertyMap.get(v)

	def getStringByPropertyType(self, v):
		return self._PropertyReverseMap[v]

	@staticmethod
	def _getStringByPropertyType(v):
		return Property._PropertyReverseMap[v]

	@staticmethod
	def _getPropertyTypeByString(v):
		return Property._PropertyMap.get(v)


# -*- coding: utf-8 -*-
# @Time    : 2023/11/28 11:50
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : Graph.py
from typing import List

from ultipa.structs.Node import Node
from ultipa.structs.Edge import Edge
from ultipa.structs.BaseModel import BaseModel

class Graph:
	node_table:List[Node] = None
	edge_table:List[Edge] = None
	def __init__(self, node_table, edge_table):
		self.node_table = node_table
		self.edge_table = edge_table


class GraphAlias(BaseModel):
	alias:str=None
	graph:Graph=None

	def __init__(self, alias, graph):
		self.alias = alias
		self.graph = graph


# -*-coding:utf-8-*-
import uuid
import json
from typing import List, Dict

from ultipa.structs.Algo import Algo, AlgoResultOpt, ALGO_RESULT
from ultipa.structs.BaseModel import BaseModel
from ultipa.structs.Edge import Edge
from ultipa.structs.EntityRow import EntityRow
from ultipa.structs.Node import Node, NodeAlias
from ultipa.structs.Path import Path, PathAlias
from ultipa.structs.Property import Property
from ultipa.structs.PropertyType import *
from ultipa.structs.ResultType import ResultType
from ultipa.structs.Retry import Retry
from ultipa.structs.Schema import Schema
from ultipa.structs.GraphSet import GraphSet
from ultipa.utils.ResposeFormat import ResponseKeyFormat
from ultipa.structs.Graph import Graph,GraphAlias
from ultipa.utils.convert import convertToListAnyObject, convertTableToDict, convertToAnyObject
from ultipa.utils.errors import ParameterException



# class TruncateType:
# 	NODES = 'nodes'
# 	EDGES = 'edges'


class DirectionType:
	left = 'left'
	right = 'right'



class TaskStatus:
	TASK_WAITING = 0
	TASK_COMPUTING = 1
	TASK_WRITEBACKING = 2
	TASK_DONE = 3
	TASK_FAILED = 4
	TASK_STOP = 5


TaskStatusString = {
	TaskStatus.TASK_WAITING: "TASK_WAITING",
	TaskStatus.TASK_COMPUTING: "TASK_COMPUTING",
	TaskStatus.TASK_WRITEBACKING: "TASK_WRITEBACKING",
	TaskStatus.TASK_DONE: "TASK_DONE",
	TaskStatus.TASK_FAILED: "TASK_FAILED",
	TaskStatus.TASK_STOP: "TASK_STOP"
}


class Code:
	SUCCESS = ultipa_pb2.SUCCESS
	FAILED = ultipa_pb2.FAILED
	PARAM_ERROR = ultipa_pb2.PARAM_ERROR
	BASE_DB_ERROR = ultipa_pb2.BASE_DB_ERROR
	ENGINE_ERROR = ultipa_pb2.ENGINE_ERROR
	SYSTEM_ERROR = ultipa_pb2.SYSTEM_ERROR
	RAFT_REDIRECT = ultipa_pb2.RAFT_REDIRECT
	RAFT_LEADER_NOT_YET_ELECTED = ultipa_pb2.RAFT_LEADER_NOT_YET_ELECTED
	RAFT_LOG_ERROR = ultipa_pb2.RAFT_LOG_ERROR
	UQL_ERROR = ultipa_pb2.UQL_ERROR
	NOT_RAFT_MODE = ultipa_pb2.NOT_RAFT_MODE
	RAFT_NO_AVAILABLE_FOLLOWERS = ultipa_pb2.RAFT_NO_AVAILABLE_FOLLOWERS
	RAFT_NO_AVAILABLE_ALGO_SERVERS = ultipa_pb2.RAFT_NO_AVAILABLE_ALGO_SERVERS
	PERMISSION_DENIED = ultipa_pb2.PERMISSION_DENIED

	UNKNOW_ERROR = 1000


class FollowerRole:
	ROLE_UNSET = ultipa_pb2.ROLE_UNSET
	ROLE_READABLE = ultipa_pb2.ROLE_READABLE
	ROLE_ALGO_EXECUTABLE = ultipa_pb2.ROLE_ALGO_EXECUTABLE




class RaftPeerInfo:
	def __init__(self, host, status=None, isLeader=False, isAlgoExecutable=False, isFollowerReadable=False,
				 isUnset=False):
		self.host = host
		self.status = status
		self.isLeader = isLeader
		self.isAlgoExecutable = isAlgoExecutable
		self.isFollowerReadable = isFollowerReadable
		self.isUnset = isUnset


class ClusterInfo:
	def __init__(self, redirect: str, raftPeers: List[RaftPeerInfo], leader: RaftPeerInfo = None):
		self.redirect = redirect
		self.leader = leader
		self.raftPeers = raftPeers


class Status:
	def __init__(self, code: Code, message: str, clusterInfo: ClusterInfo = None):
		self.code = code
		self.message = message.strip('\n')
		if clusterInfo:
			self.clusterInfo = clusterInfo




class NodeEntityTable:
	def __init__(self, schemas: List[object], nodeRows: List[EntityRow] = None):
		self.schemas = schemas
		if nodeRows is None:
			nodeRows = []
		self.nodeRows = nodeRows

	def __del__(self):
		pass



class EdgeEntityTable:
	def __init__(self, schemas: List[object], edgeRows: List[EntityRow] = None):
		self.schemas = schemas
		if edgeRows is None:
			edgeRows = []
		self.edgeRows = edgeRows

	def __del__(self):
		pass


class EdgeAlias:
	def __init__(self, alias: str, edges: List[Edge]):
		self.alias = alias
		self.edges = edges


class Attr:
	def __init__(self, alias: str, values: any, type: ResultType = None, type_desc: str = None):
		self.name = alias
		self.values = values
		self.type = type
		self.type_desc = type_desc


class AttrNode:
	def __init__(self, alias: str, values: List[List[Node]], type: ResultType = None, type_desc: str = None):
		self.name = alias
		self.values = values
		self.type = type
		self.type_desc = type_desc


class AttrEdge:
	def __init__(self, alias: str, values: List[List[Edge]], type: ResultType = None, type_desc: str = None):
		self.name = alias
		self.values = values
		self.type = type
		self.type_desc = type_desc


class AttrPath:
	def __init__(self, alias: str, values: List[List[Path]], type: ResultType = None, type_desc: str = None):
		self.name = alias
		self.values = values
		self.type = type
		self.type_desc = type_desc


class UltipaAttr:

	def __init__(self, type: PropertyType, values: any, has_attr_data: bool = False, has_ultipa_data: bool = False,
				 type_desc: any = None):
		self.values = values
		self.type = type
		self.type_desc = type_desc
		self.has_attr_data = has_attr_data
		self.has_ultipa_data = has_ultipa_data


class AttrNewAlias:
	def __init__(self, alias: str, attr: UltipaAttr):
		self.alias = alias
		self.attr = attr


class ResultAlias:
	def __init__(self, alias: str, result_type: int):
		self.alias = alias
		self.result_type = result_type


class Table(BaseModel):
	def __init__(self, table_name: str, headers: List[dict], table_rows: List[List]):
		self.name = table_name
		self.rows = table_rows
		self.headers = headers

	def getHeaders(self):
		return self.headers

	def getRows(self):
		return self.rows

	def getName(self):
		return self.name

	def headerToDicts(self) -> List[Dict]:
		return convertTableToDict(self.rows, self.headers)


class ArrayAlias:
	def __init__(self, alias: str, elements):
		self.alias = alias
		self.elements = elements


class ExplainPlan:
	def __init__(self, alias, childrenNum, uql, infos):
		self.alias = alias
		self.children_num = childrenNum
		self.uql = uql
		self.infos = infos
		self.id = str(uuid.uuid4())


class DataItem(BaseModel):

	def __init__(self, alias: str, data: any, type: str):
		self.alias = alias
		self.data = data
		self.type = type

	def asNodes(self) -> List[Node]:
		if self.type == ResultType.getTypeStr(ResultType.RESULT_TYPE_UNSET):
			if self.data is None:
				return []
			return self.data
		if self.type != ResultType.getTypeStr(ResultType.RESULT_TYPE_NODE):
			error = f"DataItem {self.alias} is not Type Node"
			raise ParameterException(error)
		return self.data

	def asFirstNodes(self) -> Node:
		if self.type == ResultType.getTypeStr(ResultType.RESULT_TYPE_UNSET):
			return self.data
		if self.type != ResultType.getTypeStr(ResultType.RESULT_TYPE_NODE):
			error = f"DataItem {self.alias} is not Type Node"
			raise ParameterException(error)
		return self.data[0] if len(self.data) > 0 else None

	def asEdges(self) -> List[Edge]:
		if self.type == ResultType.getTypeStr(ResultType.RESULT_TYPE_UNSET):
			if self.data is None:
				return []
			return self.data
		if self.type != ResultType.getTypeStr(ResultType.RESULT_TYPE_EDGE):
			error = f"DataItem {self.alias} is not Type Edge"
			raise ParameterException(error)
		return self.data

	def asFirstEdges(self) -> Edge:
		if self.type == ResultType.getTypeStr(ResultType.RESULT_TYPE_UNSET):
			return self.data
		if self.type != ResultType.getTypeStr(ResultType.RESULT_TYPE_EDGE):
			error = f"DataItem {self.alias} is not Type Edge"
			raise ParameterException(error)
		return self.data[0] if len(self.data) > 0 else None

	def asPaths(self) -> List[Path]:
		if self.type == ResultType.getTypeStr(ResultType.RESULT_TYPE_UNSET):
			if self.data is None:
				return []
			return self.data
		if self.type != ResultType.getTypeStr(ResultType.RESULT_TYPE_PATH):
			error = f"DataItem {self.alias} is not Type Path"
			raise ParameterException(error)
		return self.data

	def asAttr(self) -> Attr:
		if self.type == ResultType.getTypeStr(ResultType.RESULT_TYPE_UNSET):
			return self.data
		if self.type != ResultType.getTypeStr(ResultType.RESULT_TYPE_ATTR):
			error = f"DataItem {self.alias} is not Type Attribute list"
			raise ParameterException(error)

		return self.data

	def asGraph(self) -> Graph:
		if self.type == ResultType.getTypeStr(ResultType.RESULT_TYPE_UNSET):
			if self.data is None:
				return []
			return self.data
		if self.type != ResultType.getTypeStr(ResultType.RESULT_TYPE_GRAPH):
			error = f"DataItem {self.alias} is not Type Node"
			raise ParameterException(error)
		return self.data

	def asNodeList(self) -> AttrNode:
		return self.asAttr()

	def asEdgeList(self) -> AttrEdge:
		return self.asAttr()

	def asPathList(self) -> AttrPath:
		return self.asAttr()

	def asTable(self) -> Table:
		if self.type == ResultType.getTypeStr(ResultType.RESULT_TYPE_UNSET):
			return self.data
		if self.type != ResultType.getTypeStr(ResultType.RESULT_TYPE_TABLE):
			error = f"DataItem {self.alias} is not Type Table"
			raise ParameterException(error)
		return self.data

	def asSchemas(self) -> List[Schema]:
		if self.type == ResultType.getTypeStr(ResultType.RESULT_TYPE_UNSET):
			if self.data is None:
				return []
			return self.data
		if self.type != ResultType.getTypeStr(ResultType.RESULT_TYPE_TABLE):
			error = f"DataItem {self.alias} is not Type Table"
			raise ParameterException(error)
		alias = self.data.getName()
		if alias.startswith("_node"):
			type = "node"
		elif alias.startswith("_edge"):
			type = "edge"

		else:
			type = None

		headers = self.data.getHeaders()
		rows = self.data.getRows()
		tableListDict = convertTableToDict(rows, headers)
		REPLACE_KEYS = {
			"totalNodes": "total",
			"totalEdges": "total",
		}
		BOOL_KEYS = ["index", "lte"]
		JSON_KEYS = ["properties"]
		convert2Int = ["totalNodes", "totalEdges"]
		responseKeyFormat = ResponseKeyFormat(keyReplace=REPLACE_KEYS, boolKeys=BOOL_KEYS, jsonKeys=JSON_KEYS,
											  convert2Int=convert2Int)
		dataList = responseKeyFormat.changeKeyValue(tableListDict)
		schemaList = []
		for data in dataList:
			responseKeyFormat = ResponseKeyFormat(boolKeys=BOOL_KEYS)
			properList = responseKeyFormat.changeKeyValue(data.get("properties"))
			propertyList = [
				Property(name=propo.get("name"), description=propo.get("description"), type=propo.get("type"),
						 lte=propo.get("lte")) for propo in properList]

			schemaList.append(
				Schema(name=data.get("name"), description=data.get("description"), properties=propertyList,
					   dbType=None, type=type, total=data.get("total")))

		return schemaList

	def asProperties(self) -> List[Property]:
		if self.type == ResultType.getTypeStr(ResultType.RESULT_TYPE_UNSET):
			if self.data is None:
				return []
			return self.data
		if self.type != ResultType.getTypeStr(ResultType.RESULT_TYPE_TABLE):
			error = f"DataItem {self.alias} is not Type Table"
			raise ParameterException(error)
		headers = self.data.getHeaders()
		rows = self.data.getRows()
		table_rows_dict = convertTableToDict(rows, headers)
		BOOL_KEYS = ["lte"]
		responseKeyFormat = ResponseKeyFormat(boolKeys=BOOL_KEYS)
		dataList = responseKeyFormat.changeKeyValue(table_rows_dict)
		return [Property(name=data.get("name"), description=data.get("description"), type=data.get("type"),
						 lte=data.get("lte"), schema=data.get("schema")) for data in dataList]

	def asGraphInfos(self) -> List[GraphSet]:
		REPLACE_KEYS = {
			"graph": "name",
		}
		if self.type == ResultType.getTypeStr(ResultType.RESULT_TYPE_UNSET):
			if self.data is None:
				return []
			return self.data
		if self.type != ResultType.getTypeStr(ResultType.RESULT_TYPE_TABLE):
			error = f"DataItem {self.alias} is not Type Table"
			raise ParameterException(error)
		headers = self.data.getHeaders()
		rows = self.data.getRows()
		table_rows_dict = convertTableToDict(rows, headers)
		responseKeyFormat = ResponseKeyFormat(keyReplace=REPLACE_KEYS)
		data = responseKeyFormat.changeKeyValue(table_rows_dict)
		data = convertToListAnyObject(data)
		return data

	def asAlgos(self) -> List[Algo]:
		if self.type == ResultType.getTypeStr(ResultType.RESULT_TYPE_UNSET):
			if self.data is None:
				return []
			return self.data
		if self.type != ResultType.getTypeStr(ResultType.RESULT_TYPE_TABLE):
			error = f"DataItem {self.alias} is not Type Table"
			raise ParameterException(error)
		headers = self.data.getHeaders()
		rows = self.data.getRows()
		table_rows_dict = convertTableToDict(rows, headers)
		algo_list = []
		for data in table_rows_dict:
			paramDict = json.loads(data.get("param"))
			result_opt = int(paramDict.get("result_opt"))
			paramDict.update(paramDict)
			result_opt_obj = AlgoResultOpt()
			result_opt_obj.can_realtime = True if result_opt & ALGO_RESULT.WRITE_TO_CLIENT else False
			result_opt_obj.can_visualization = True if result_opt & ALGO_RESULT.WRITE_TO_VISUALIZATION else False
			result_opt_obj.can_write_back = True if result_opt & (
					ALGO_RESULT.WRITE_TO_DB | ALGO_RESULT.WRITE_TO_FILE) else False
			paramDict.update({"result_opt": result_opt_obj})
			algo_list.append(paramDict)

			algoObj = Algo(paramDict.get("name"), paramDict.get("description"), paramDict.get("version"),
						   paramDict.get("result_opt"),
						   paramDict.get("parameters"), paramDict.get("write_to_stats_parameters"),
						   paramDict.get("'write_to_db_parameters'"),
						   paramDict.get("'write_to_file_parameters'"))

			algo_list.append(algoObj)
		return algo_list

	def asAny(self) -> any:
		return self.data

	def asKV(self):
		return self.toDict()


class BaseUqlReply:
	def __init__(self, paths: List[PathAlias], nodes: List[NodeAlias], edges: List[EdgeAlias], tables: List[Table],
				 graphs:List[GraphAlias],attrs: List = None, resultAlias: List = None,explainPlan: List[ExplainPlan] = None):
		self.paths = paths
		self.nodes = nodes
		self.edges = edges
		self.tables = tables
		self.attrs = attrs
		self.graphs = graphs
		# self.arrays = arrays
		self.resultAlias = resultAlias
		self.explainPlan = explainPlan


class UltipaStatistics(BaseModel):
	def __init__(self, edge_affected: int, node_affected: int, engine_time_cost: int, total_time_cost: int):
		self.edgeAffected = edge_affected
		self.nodeAffected = node_affected
		self.engineCost = engine_time_cost
		self.totalCost = total_time_cost


class UqlReply(BaseModel):
	datas: List[DataItem]

	def __init__(self, dataBase: BaseUqlReply, aliasMap: dict = None, datas: List = None):
		if aliasMap == None:
			aliasMap = {}
		self._aliasMap = aliasMap
		if datas is None:
			datas = []
		self.datas: List[DataItem] = datas
		self.explainPlan: List[ExplainPlan] = []
		self._dataBase = dataBase

		for data in self._dataBase.paths:
			if self._aliasMap.get(data.alias):
				self._aliasMap[data.alias].data.extend(data.paths)
				continue
			self._aliasMap[data.alias] = DataItem(data.alias, data.paths,
												  ResultType.getTypeStr(ResultType.RESULT_TYPE_PATH))


		for data in self._dataBase.graphs:
			if self._aliasMap.get(data.alias):
				self._aliasMap[data.alias].data.extend(data.graph)
				continue
			self._aliasMap[data.alias] = DataItem(data.alias, data.graph,
												  ResultType.getTypeStr(ResultType.RESULT_TYPE_GRAPH))

		for data in self._dataBase.nodes:
			if self._aliasMap.get(data.alias):
				self._aliasMap[data.alias].data.extend(data.nodes)
				continue
			self._aliasMap[data.alias] = DataItem(data.alias, data.nodes,
												  ResultType.getTypeStr(ResultType.RESULT_TYPE_NODE))

		for data in self._dataBase.edges:
			if self._aliasMap.get(data.alias):
				self._aliasMap[data.alias].data.extend(data.edges)
				continue
			self._aliasMap[data.alias] = DataItem(data.alias, data.edges,
												  ResultType.getTypeStr(ResultType.RESULT_TYPE_EDGE))

		for data in self._dataBase.attrs:
			if self._aliasMap.get(data.name):
				self._aliasMap[data.name].data.append(data)
				continue
			self._aliasMap[data.name] = DataItem(data.name, data, ResultType.getTypeStr(ResultType.RESULT_TYPE_ATTR))

		for data in self._dataBase.tables:
			if self._aliasMap.get(data.name):
				self._aliasMap[data.name].data.extend(data)
				continue
			self._aliasMap[data.name] = DataItem(data.name, data, ResultType.getTypeStr(ResultType.RESULT_TYPE_TABLE))

		for data in self._dataBase.explainPlan:
			self.explainPlan.append(data)

		for data in self._dataBase.resultAlias:
			if self._aliasMap.get(data.alias):
				self.datas.append(self._aliasMap[data.alias])
		if not self.datas:
			for key in self._aliasMap:
				self.datas.append(self._aliasMap[key])


class ReturnReq:
	def __init__(self, graphSetName: str, uql: str, host: str, retry: Retry, uqlIsExtra: bool):
		self.graph_name = graphSetName
		self.uql = uql
		self.host = host
		self.Retry = retry
		self.uqlIsExtra = uqlIsExtra


class ExportReply:
	def __init__(self, data: List[NodeAlias]):
		self.data = data


class PaserAttrListData:
	def __init__(self, type, nodes: List[Node] = None, edges: List[Edge] = None, paths: List[Path] = None,
				 attrs: UltipaAttr = None):
		self.type = type
		self.nodes = nodes
		self.edges = edges
		self.paths = paths
		self.attrs = attrs


import copy

from ultipa.structs.Retry import Retry, RetryResponse
from ultipa.types import ULTIPA, ULTIPA_REQUEST


class RetryHelp:
	'''
	    Configuration class that defines settings for the retry mechanism.
	'''
	currentRetry: Retry = None

	@staticmethod
	def getRty(requestConfig):
		if requestConfig.retry:
			nextRetry = requestConfig.retry
		else:
			nextRetry = Retry(current=0, max=3)

		if not RetryHelp.currentRetry:
			RetryHelp.currentRetry = copy.deepcopy(nextRetry)
		return RetryHelp.currentRetry, nextRetry

	@staticmethod
	def check(conn, requestConfig, response) -> RetryResponse:
		canRetry: bool = False
		currentRetry, nextRetry = RetryHelp.getRty(requestConfig)
		if response.status.code in [
			ULTIPA.Code.RAFT_REDIRECT,
			ULTIPA.Code.RAFT_LEADER_NOT_YET_ELECTED,
			ULTIPA.Code.RAFT_NO_AVAILABLE_FOLLOWERS,
			ULTIPA.Code.RAFT_NO_AVAILABLE_ALGO_SERVERS
		]:
			if nextRetry.current < nextRetry.max:
				redirectHost = response.status.clusterInfo.redirect
				refresh = conn.refreshRaftLeader(redirectHost, requestConfig)
				if refresh:
					nextRetry.current += 1
					canRetry = True
		reTry = RetryResponse(canRetry, currentRetry, nextRetry)
		return reTry

	@staticmethod
	def checkRes(response):
		if response.status.code in [
			ULTIPA.Code.RAFT_REDIRECT,
			ULTIPA.Code.RAFT_LEADER_NOT_YET_ELECTED,
			ULTIPA.Code.RAFT_NO_AVAILABLE_FOLLOWERS,
			ULTIPA.Code.RAFT_NO_AVAILABLE_ALGO_SERVERS
		]:
			return True
		else:
			return False



# -*- coding: utf-8 -*-
# @Time    : 2023/8/1 11:47
# @Author  : Ultipa
# @Email   : support@ultipa.com
# @File    : Schema.py
from typing import List

from ultipa.structs import DBType
from ultipa.structs.Property import Property
from ultipa.structs.BaseModel import BaseModel


class Schema(BaseModel):
	'''
	    Data class for schema.
	'''

	def __init__(self, name: str, dbType: DBType, type: str = None, description: str = None,
				 properties: List[Property] = None,
				 total: int = None):
		self.description = description
		self.properties = properties
		self.name = name
		self.total = total
		self.type = type
		self.DBType = dbType

	def getProperty(self, name: str):
		find = list(filter(lambda x: x.get('name') == name, self.properties))
		if find:
			return find[0]
		return None
